from threeML.models.tablemodel import FITSTableModel
import numpy
import collections
from threeML.models.Parameter import Parameter
import math
import scipy.integrate
import operator



class Pion(FITSTableModel):

    def setup(self):
        
        self.SetTableFile("/Users/jburgess/Research/3ML/threeml/models/fluxModels/MyTemplates/piontemplate.fits")

                
        self.functionName        = "Solar Flare Pion"
        self.formula             = r"Ask Nicola"
        self.parameters          = collections.OrderedDict()
        self.parameters['S'] = Parameter('S',3.,2.,6.,0.1,fixed=False,nuisance=False,dataset=None)

        self.integral = 1.        
    
        self.ncalls              = 0
    
        
 
  
    def __call__(self,energy):
        self.ncalls             += 1
        S                       = self.parameters['S'].value
        
     
        return self._interpFunc((S,energy))



import warnings as custom_warnings


# Monkeypatch the print of warning so we can customize them

def my_format_warning(message, category, *args):
    """
    Override the default showwarning to customize the appearance of warnings
    :return:
    """

    return "\nWARNING %s: %s\n\n" % (category.__name__, message)


custom_warnings.formatwarning = my_format_warning

class ModelAssertionViolation(Exception):

    pass


class CppInterfaceNotAvailable(Warning):

    pass

class CannotImportPlugin(Warning):

    pass


#Implements a minimal reader for OGIP PHA format for spectral data
# (https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/spectra/ogip_92_007/node6.html)

#Author: Giacomo Vianello (giacomov@slac.stanford.edu)

import os
import astropy.io.fits as pyfits
import numpy

requiredKeywords             = {}
requiredKeywords['observed'] = ("mission:TELESCOP,instrument:INSTRUME,filter:FILTER,"+
                                 "exposure:EXPOSURE,backfile:BACKFILE,"+
                                 "corrfile:CORRFILE,corrscal:CORRSCAL,respfile:RESPFILE,"+
                                 "ancrfile:ANCRFILE,hduclass:HDUCLASS,"+
                                 "hduclas1:HDUCLAS1,hduvers:HDUVERS,poisserr:POISSERR,"+
                                 "chantype:CHANTYPE,n_channels:DETCHANS").split(",")
requiredKeywords['background'] = ("mission:TELESCOP,instrument:INSTRUME,filter:FILTER,"+
                                 "exposure:EXPOSURE,"+
                                 "hduclass:HDUCLASS,"+
                                 "hduclas1:HDUCLAS1,hduvers:HDUVERS,poisserr:POISSERR,"+
                                 "chantype:CHANTYPE,n_channels:DETCHANS").split(",")
mightBeColumns                = {}
mightBeColumns['observed']    = ("EXPOSURE,BACKFILE,"+
                                 "CORRFILE,CORRSCAL,"+
                                 "RESPFILE,ANCRFILE").split(",")
mightBeColumns['background']  = ("EXPOSURE").split(",")


class OGIPPHA(object):
  def __init__(self,phafile,spectrumNumber=None,**kwargs):
    
    self.filetype           = 'observed'
    for k,v in kwargs.iteritems():
      if(k.lower()=="filetype"):
        if(v.lower()=="background"):
          self.filetype       = "background"
        elif(v.lower()=="observed"):
          self.filetype       = "observed"
        else:
          raise RuntimeError("Unrecognized filetype keyword value")
      pass
    pass
    
    #Allow the use of a syntax like "mySpectrum.pha{1}" to specify the spectrum
    #number in PHA II files
    ext                     = os.path.splitext(phafile)[-1]
    if('{' in ext):
      spectrumNumber        = int(ext.split('{')[-1].replace('}',''))
      phafile               = phafile.split('{')[0]
    
    with pyfits.open(phafile) as f:
      try:
        HDUidx                = f.index_of("SPECTRUM")
      except:
        raise RuntimeError("The input file %s is not in PHA format" %(phafile))
      pass
      
      self.spectrumNumber     = spectrumNumber
      
      spectrum                = f[HDUidx]
      data                    = spectrum.data
      header                  = spectrum.header
      
      #Determine if this file contains COUNTS or RATES
      if("COUNTS" in data.columns.names):
        self.hasRates         = False
        self.dataColumnName   = "COUNTS"
      elif("RATE" in data.columns.names):
        self.hasRates         = True
        self.dataColumnName   = "RATE"
      else:
        raise RuntimeError("This file does not contain a RATE nor a COUNTS column. This is not a valid PHA file")
      pass
      
      #Determine if this is a PHA I or PHA II
      if(len(data.field(self.dataColumnName).shape)==2):
        self.typeII           = True
        if(self.spectrumNumber==None):
          raise RuntimeError("This is a PHA Type II file. You have to provide a spectrum number")
      else:
        self.typeII           = False
      pass
      
      #Collect informations from mandatory keywords
      keys                    = requiredKeywords[self.filetype]
      for k in keys:
        internalName,keyname  = k.split(":")
        if(keyname in header):
          self.__setattr__(internalName,header[keyname])
        else:
          if(keyname in mightBeColumns[self.filetype] and self.typeII):
            #Check if there is a column with this name
            if(keyname in data.columns.names):
              self.__setattr__(internalName,data[keyname][self.spectrumNumber-1])
            else:
              raise RuntimeError("Keyword %s is not in the header nor in the data extension. This file is not a proper PHA file" % keyname)
          else:
            #The keyword POISSERR is a special case, because even if it is missing,
            #it is assumed to be False if there is a STAT_ERR column in the file
            if(keyname=="POISSERR" and "STAT_ERR" in data.columns.names):
              self.poisserr   = False
            else:
              raise RuntimeError("Keyword %s not found. File %s is not a proper PHA file" % (keyname,phafile))
            pass
          pass
        pass
      pass
      
      #Now get the data (counts or rates) and their errors. If counts, transform them in rates
      if(self.typeII):
      
        #PHA II file
        if(self.hasRates):
          self.rates          = data.field(self.dataColumnName)[self.spectrumNumber-1,:]
          if(not self.poisserr):
            self.ratesErrors  = data.field("STAT_ERR")[self.spectrumNumber-1,:]
          pass
        else:
          self.rates          = data.field(self.dataColumnName)[self.spectrumNumber-1,:]/self.exposure
          if(not self.poisserr):
            self.ratesErrors  = data.field("STAT_ERR")[self.spectrumNumber-1,:]/self.exposure
        pass
        
        if("SYS_ERR" in data.columns.names):
            self.sysErrors    = data.field("SYS_ERR")[self.spectrumNumber-1,:]
        else:
            self.sysErrors    = numpy.zeros(self.rates.shape)
        pass
      
      elif(self.typeII==False):
        
        #PHA 1 file
        if(self.hasRates):
          self.rates          = data.field(self.dataColumnName)
          if(not self.poisserr):
            self.ratesErrors  = data.field("STAT_ERR")
        else:
          self.rates          = data.field(self.dataColumnName)/self.exposure
          if(not self.poisserr):
            self.ratesErrors  = data.field("STAT_ERR")/self.exposure
        pass
        
        if("SYS_ERR" in data.columns.names):
            self.sysErrors    = data.field("SYS_ERR")
        else:
            self.sysErrors    = numpy.zeros(self.rates.shape)
        pass
      pass
      
    pass #Closing the FITS file 
  pass
  
  def getRates(self):
    return self.rates
  
  def getRatesErrors(self):
    return self.ratesErrors
  
  def getSysErrors(self):
    return self.sysErrors
  
  def getExposure(self):
    return self.exposure
  
pass



from band import Band
from powerlaw import PowerLaw
from logparabola import LogParabola, LogParabolaEp
from madness import Madness
from manybrokenpowerlaws import ManyBrokenPowerlaws
from manylogporabolas import ManyLogParabolas
from blackbody import Blackbody
from expCutoff import ExponentialCutoff
from gaussian import Gaussian
from asaf import AsafModel
from bandTable import BandTable
from pion import Pion
from nar import NAR

from CustomSpectralModel import CustomSpectralModel



import numpy
import math

import warnings
import matplotlib.pyplot as plt
import collections

from threeML.io.Table import Table, NumericMatrix
from threeML.utils.cartesian import cartesian
from threeML.io.ProgressBar import ProgressBar

from iminuit import Minuit

from IPython.display import display

import uncertainties

import re

#Special constants
FIT_FAILED = -1e12

class Minimizer(object):
  def __init__(self,function,parameters,ftol=1e-3,verbosity=1):
    
    self.function             = function
    self.parameters           = parameters
    self.Npar                 = len(self.parameters.keys())
    self.ftol                 = ftol
    self.verbosity            = verbosity
  
  def minimize(self):
    raise NotImplemented("This is the method of the base class. Must be implemented by the actual minimizer")

#This is a function to add a method to a class
#We will need it in the MinuitMinimizer

def add_method(self, method, name=None):
    if name is None:
        name = method.func_name
    setattr(self.__class__, name, method)


class iMinuitMinimizer(Minimizer):
  def __init__(self,function,parameters,ftol=1e3,verbosity=0):
    
    super(iMinuitMinimizer, self).__init__(function,parameters,ftol,verbosity)
    
    #Prepare the parameter dictionary for iminuit
    pars                      = {}
    varnames                  = []
    
    for k,par in parameters.iteritems():
      
      curName                 = "%s_of_%s" %(k[1],k[0])
      
      varnames.append(curName)
      
      #Initial value
      pars['%s' % curName]       = par.value
      
      #Initial delta
      pars['error_%s' % curName] = par.delta
      
      #Limits
      pars['limit_%s' % curName] = ( par.minValue, par.maxValue )
      
      #This is useless, since all parameters here are free,
      #but do it anyway for clarity
      pars['fix_%s' % curName]   = False
    
    #This is to tell Minuit that we are dealing with likelihoods,
    #not chi square
    pars['errordef']          = 0.5
        
    pars['print_level']       = verbosity
    
    #We need to make a function with the parameters as explicit
    #variables in the calling sequence, so that Minuit will be able
    #to probe the parameter's names
    var_spelledout            = ",".join(varnames)
    
    #A dictionary to keep a way to convert from var. name to
    #variable position in the function calling sequence 
    #(will use this in contours)
    
    self.nameToPos = { k: i for i, k in enumerate( varnames ) }
    
    #Write and compile the code for such function
      
    code                      = 'def _f(self, %s):\n  return self.function(%s)' % (var_spelledout, var_spelledout)
    exec(code)
    
    #Add the function just created as a method of the class
    #so it will be able to use the 'self' pointer
    add_method(self, _f, "_f")    
        
    #Finally we can instance the Minuit class
    self.minuit               = Minuit(self._f, **pars)
        
    self.minuit.tol           = 100 #ftol
    
    try:
        
        self.minuit.up            = 0.5 #This is a likelihood
    
    except:
        
        #iMinuit uses errodef, not up
        
        self.minuit.errordef  = 0.5
    
    self.minuit.strategy      = 0 #More accurate
    
  def migradConverged(self):
    
    #In the MINUIT manual this is the condition for MIGRAD to have converged
    #0.002 * tolerance * UPERROR (which is 0.5 for likelihood)
    return ( self.minuit.edm <= 0.002 * self.minuit.tol * 0.5)
  
  
  def _runMigrad(self, trials=2):
    
    #Repeat Migrad up to 10 times, until it converges
    
    for i in range(trials):
      
      self.minuit.migrad()
      
      if self.migradConverged():
        
        #Converged
        break
      
      else:
        
        #Try again
        continue
  
  def minimize(self):
    
    self._runMigrad()
    
    if not self.migradConverged():
      
      print("\nMIGRAD did not converge in 10 trials.")
      
      return map(lambda x:0,self.minuit.values), 1e9
    
    else:
            
      #Make a ordered dict for the results
      bestFit                   = collections.OrderedDict()
      
      for k,par in self.parameters.iteritems():
        
        curName                 = "%s_of_%s" %(k[1],k[0])
        
        bestFit[curName]        = self.minuit.values[curName]
      
      #NOTE: hesse must be callsed AFTER the fit because it
      #will change the value of the parameters
      
      self.minuit.hesse()
      
      #Restore best fit
      for k,par in self.parameters.iteritems():
        
        curName                 = "%s_of_%s" %(k[1],k[0])
        
        par.setValue( bestFit[curName] )
      
      return bestFit, self.minuit.fval
  
  def printFitResults(self):
    
    #I do not use the print_param facility in iminuit because
    #it does not work well with console output, since it fails
    #to autoprobe that it is actually run in a console and uses
    #the HTML backend instead
    
    data = []
    nameLength = 0
    
    for k,v in self.parameters.iteritems():
     
     curName                 = "%s_of_%s" %(k[1],k[0])
     
     #Format the value and the error with sensible significant
     #numbers
     x                      = uncertainties.ufloat(v.value, self.minuit.errors[curName])
     rep                    = x.__str__().replace("+/-"," +/- ")
          
     data.append([curName,rep,v.unit])
     
     if(len(v.name) > nameLength):
       nameLength = len(curName)
   
    pass
    
    table                     = Table(rows = data,
                                      names = ["Name","Value","Unit"],
                                      dtype=('S%i' %nameLength, str, 'S15'))
    
    display(table)
    print("\nNOTE: errors on parameters are approximate. Use getErrors().\n")
      
  def printCorrelationMatrix(self):
        
    #Print a custom covariance matrix because iminuit does
    #not guess correctly the frontend when 3ML is used
    #from terminal
    
    cov                       = self.minuit.covariance
    
    if cov is None:
        
        raise RuntimeError("Cannot compute covariance. This usually means that there are unconstrained" + 
                           " parameters. Fix those or reduce their allowed range, or use a simpler model.")
    
    keys                      = self.parameters.keys()
    
    parNames                  = map(lambda k:"%s_of_%s" %(k[1],k[0]), keys ) 
    
    data = []
    nameLength = 0
    
    for key1, name1 in zip(keys, parNames):
       
      if( len(name1) > nameLength ):
        
        nameLength = len(name1)
     
      thisRow                 = []
      
      v1                      = self.parameters[key1]
      
      for key2, name2 in zip(keys, parNames):
       
        corr                  = cov[(name1, name2)] / ( math.sqrt(cov[(name1, name1)]) * math.sqrt(cov[(name2, name2)]) ) 
       
        thisRow.append(corr)
      
      pass
      
      data.append(thisRow)
      
    pass
    
    dtypes                    = []
    dtypes.extend( map(lambda x:float, parNames) )
    
    cols                      = []
    cols.extend(parNames)
    
    table                     = NumericMatrix(rows = data,
                                      names = cols,
                                      dtype= dtypes)
    
    for col in table.colnames:
      
      table[col].format = '2.2f'
    
    display(table)
      
  def getErrors(self,fast): 
    
    #Run again the fit because the user might have changed the parameter
    #configuration
    self._runMigrad()
    
    #Now set aside the current values for the parameters,
    #because minos will change them
    #Make a ordered dict for the results
    bestFit                   = collections.OrderedDict()
    
    for k,par in self.parameters.iteritems():
      
      curName                 = "%s_of_%s" %(k[1],k[0])
      
      bestFit[curName]        = self.minuit.values[curName]

    
    if not self.migradConverged():
      
      print("\nMIGRAD results are not valid. Cannot compute errors. Did you run the fit first ?")
      
      return map(lambda x:None, self.parameters.keys())
    
    try:
    
      self.minuit.minos()
    
    except:
      
      print("MINOS has failed. This usually means that the fit is very difficult, for example "
            "because of high correlation between parameters. Check the correlation matrix printed"
            "in the fit step, and check contour plots with getContours(). If you are using a "
            "user-defined model, you can also try to "
            "reformulate your model with less correlated parameters.")
      
      return None
          
    #Make a ordered dict for the results
    errors                   = collections.OrderedDict()
    
    for k,par in self.parameters.iteritems():
      
      curName                 = "%s_of_%s" %(k[1],k[0])
      
      errors[curName]         = (self.minuit.merrors[(curName, -1)], self.minuit.merrors[(curName, 1)])
      
      #Set the parameter back to the best fit value
      par.setValue( bestFit[curName] )
    
    
    #Print a table with the errors
    data = []
    nameLength = 0
    
    for k,v in self.parameters.iteritems():
    
      curName                 = "%s_of_%s" %(k[1],k[0])
      
      #Format the value and the error with sensible significant
      #numbers
      x                      = uncertainties.ufloat(v.value, abs(errors[curName][0]))
      
      num, uncm, ex          = re.match('\(?(\-?[0-9]+\.?[0-9]+) ([0-9]+\.[0-9]+)\)?(e[\+|\-][0-9]+)?',
                                        x.__str__().replace("+/-"," ")).groups()
      
      x                      = uncertainties.ufloat(v.value, abs(errors[curName][1]))
      _, uncp, _             = re.match('\(?(\-?[0-9]+\.?[0-9]+) ([0-9]+\.[0-9]+)\)?(e[\+|\-][0-9]+)?',
                                        x.__str__().replace("+/-"," ")).groups()
      
      if(ex==None):
        
        #Number without exponent
        prettystring           = "%s -%s +%s" %(num, uncm, uncp)
      
      else:
        
        prettystring           = "(%s -%s +%s)%s" %(num, uncm, uncp, ex)
        
      data.append([curName,prettystring,v.unit])
     
      if(len(v.name) > nameLength):
        nameLength = len(curName)
    pass
    
    table                     = Table(rows = data,
                                      names = ["Name","Value","Unit"],
                                      dtype=('S%i' %nameLength, str, 'S15'))
    
    display(table)

    
    return errors
  
  def contours(self, src1, param1, p1min, p1max, p1steps,
                    src2, param2, p2min, p2max, p2steps,
                    progress=True, **kwargs):
    
    #First create another minimizer 
    
    newargs = dict( self.minuit.fitarg )
    
    #Update the values for the parameters with the best fit one
    
    for key, value in self.minuit.values.iteritems():
        
        newargs[key] = value
    
    #Fix the parameters under scrutiny
    
    #values = {}
    
    for s,p in zip( [src1, src2], [param1,param2] ):
        
        if s is None:
            
            #Only one parameter to analyze
            
            continue
        
        key = "%s_of_%s" %(p,s)
        
        if key not in newargs.keys():
            
            raise ValueError("Parameter %s is not a free parameter for source %s." %(p,s))
        
        else:
            
            newargs[ 'fix_%s' % key ] = True
        
        #values[ key ] = float( self.minuit.values[key] )
    
    #This is a likelihood
    newargs['errordef'] = 0.5
    
    newargs['print_level'] = 0
        
    #Now create the new minimizer
    self.contour_minuit = Minuit( self._f, **newargs )
    
    #Check the keywords
    p1log = False
    p2log = False
    
    if 'log' in kwargs.keys():
        
        p1log = bool( kwargs['log'][0] )
        
        if param2 is not None:
            
            p2log = bool( kwargs['log'][1] )
    
    #Generate the steps
    
    if p1log:
    
        a = numpy.logspace( numpy.log10(p1min), numpy.log10(p1max), p1steps)
    
    else:
        
        
        a = numpy.linspace( p1min, p1max, p1steps)
    
    if param2 is not None:
    
        if p2log:
        
            b = numpy.logspace( numpy.log10(p2min), numpy.log10(p2max), p2steps)
        
        else:
        
            b = numpy.linspace( p2min, p2max, p2steps)
    
    else:
        
        #Only one parameter to step through
        #Put b as nan so that the worker can realize that it does not have
        #to step through it
        
        b = numpy.array([numpy.nan])
    
    #Generate the grid
    
    grid = cartesian([a,b])
    
    #Define the parallel worker
    
    def contourWorker(args):
            
      aa, bb = args
      
      name1 = "%s_of_%s" % ( param1, src1 )
      
      #Will change this if needed
      name2 = None
      
      #First of all restore the best fit values
      #for k,v in values.iteritems():
          
      #    self.minuit.values[ k ] = v
      
      #Now set the parameters under scrutiny to the current values
      
      #Since iminuit does not allow to fix parameters,
      #I am forced to create a new one (which sucks)
      
      newargs = dict( self.minuit.fitarg )
      
      newargs[ 'fix_%s' % name1 ] = True
      newargs[ '%s' % name1 ] = aa
            
      if numpy.isfinite(bb):
          
          name2 = "%s_of_%s" % ( param2, src2 )
          
          newargs[ 'fix_%s' % name2 ] = True
          newargs[ '%s' % name2 ] = bb
      
      else:
          
          #We are stepping through one param only.
          #Do nothing
          
          pass
      
      newargs['errordef'] = 0.5
      newargs['print_level'] = 0
      
      m = Minuit(self._f, **newargs )
      
      #High tolerance for speed
      m.tol = 100
      
      #mpl.warning("Running migrad")
      
      #Handle the corner case where there are no free parameters
      #after fixing the two under scrutiny
      
      #free = [k for k, v in self.contour_minuit.fixed.iteritems() if not v]
            
      if len( m.list_of_vary_param() )==0:
          
          #All parameters are fixed, just return the likelihood function
          
          if name2 is None:
              
              val = self._f(aa)
          
          else:
              
              #This is needed because the user could specify the
              #variables in reverse order
              
              myvars = [0] * 2
              myvars[ self.nameToPos[ name1 ] ] = aa
              myvars[ self.nameToPos[ name2 ] ] = bb 
          
              val = self._f(*myvars)
                    
          return val
          

      try:
        
        m.migrad()
      
      except:
            
        #In this context this is not such a big deal,
        #because we might be so far from the minimum that
        #the fit cannot converge
        
        return FIT_FAILED
      
      else:
        
        pass
        
        #print("%s -> %s" % (self.minuit.values, self.minuit.fval))
      
      #mpl.warning("Returning")
            
      return m.fval    
    
    #Do the computation
    
    if(progress):
      
      prog = ProgressBar(grid.shape[0])
      
      def wrap(args):
        prog.increase()
        return contourWorker(args)
      
      r = map( wrap, grid )
      
    else:
      
      r = map( contourWorker, grid )
    
    return a,b,numpy.array(r).reshape((a.shape[0],b.shape[0]))



from SourceModel import SourceModel
import collections
from Parameter import Parameter

class PointSource(SourceModel):
  
  def __init__(self,name,ra,dec,spectralModel):
    
    #TODO: accept input coordinates in other systems and convert them immediately
    #to J2000 Equatorial R.A. and Dec.
    
    self.name                 = name
    self.functionName         = "Point source"
    self.formula              = r"\begin{equation}f(RA',Dec') = \delta(RA'-RA)\delta(Dec'-Dec)\end{equation}"
    self.parameters           = collections.OrderedDict()
    self.parameters['RA']     = Parameter('RA', ra, 0.0, 360.0, 0.01,fixed=True,nuisance=False,dataset=None,unit='deg')
    self.parameters['Dec']    = Parameter('Dec', dec, -90.0, 90.0, 0.01,fixed=True,nuisance=False,dataset=None,unit='deg')
    
    if(not callable(spectralModel)):
      
      raise RuntimeError("The provided spectral model must be callable")
    
    else:
    
      self.spectralModel      = spectralModel
    
    pass
  pass    
  
  def getFlux(self,energies):
    return self.spectralModel(energies)
  
  def getPosition(self):
    return (self.parameters['RA'].value, self.parameters['Dec'].value)
  
pass




#!/usr/bin/python
# angsep.py
# Program to calculate the angular separation between two points
# whose coordinates are given in RA and Dec
# From angsep.py Written by Enno Middelberg 2001

from numpy import *
import string
import sys

def angsep(ra1deg,dec1deg,ra2deg,dec2deg):
    """ Determine separation in degrees between two celestial objects
        arguments are RA and Dec in decimal degrees.
        """
    ra1rad=ra1deg*pi/180
    dec1rad=dec1deg*pi/180
    ra2rad=ra2deg*pi/180
    dec2rad=dec2deg*pi/180
    
    # calculate scalar product for determination
    # of angular separation
    
    x=cos(ra1rad)*cos(dec1rad)*cos(ra2rad)*cos(dec2rad)
    y=sin(ra1rad)*cos(dec1rad)*sin(ra2rad)*cos(dec2rad)
    z=sin(dec1rad)*sin(dec2rad)
    
    rad=arccos(x+y+z) # Sometimes gives warnings when coords match
    
    # use Pythargoras approximation if rad < 1 arcsec
    sep = choose( rad<0.000004848 , (sqrt((cos(dec1rad)*(ra1rad-ra2rad))**2+(dec1rad-dec2rad)**2),rad))
        
    # Angular separation
    sep=sep*180/pi
                                     
    return sep


from threeML.parallel.ParallelClient import ParallelClient

from threeML.config.Config import threeML_config

from threeML.io.ProgressBar import ProgressBar

from threeML.io.triangle import corner

from threeML.exceptions.CustomExceptions import ModelAssertionViolation

import emcee
import emcee.utils

import numpy
import time

def sampleWithProgress( p0, sampler, nsamples, **kwargs ):
    
    progress = ProgressBar( nsamples )
    
    for i, result in enumerate( sampler.sample(p0, iterations = nsamples, **kwargs) ):
        
        progress.animate( ( i + 1 ) )
        
        pos, prob, state = result
    
    progress.animate( nsamples )
    print("")
    return pos, prob, state

def sampleWithoutProgress( p0, sampler, nsamples, **kwargs ):
    
    return sampler.run_mcmc( p0, nsamples, **kwargs )

class bayesianAnalysis( object ):
    
    def __init__(self, likelihoodModel, dataList, **kwargs):
    
        #Process optional keyword parameters
        self.verbose             = False
            
        for k,v in kwargs.iteritems():
          
          if(k.lower()=="verbose"):
          
            self.verbose           = bool(kwargs["verbose"])
        
            
        self.likelihoodModel      = likelihoodModel
            
        self.dataList             = dataList
        
        for dataset in self.dataList.values():
          
          dataset.setModel(self.likelihoodModel)
    
    def _logp( self, trialValues ):
        
        #Compute the sum of the log-priors
        
        logp = 0
        
        for i,( srcName, paramName ) in enumerate( self.freeParameters.keys() ):
            
            thisParam = self.likelihoodModel.parameters[ srcName ][ paramName ]
            
            logp += thisParam.prior( trialValues[i] )
            
        return logp
        
    def _logLike( self, trialValues ):
        
        #Compute the log-likelihood
                
        for i,( srcName, paramName ) in enumerate( self.freeParameters.keys() ):
            
            thisParam = self.likelihoodModel.parameters[ srcName ][ paramName ]
            
            thisParam.setValue( trialValues[i] )
        
        try:    
            
            logLike = numpy.sum( map( lambda dataset: dataset.getLogLike(), self.dataList.values() ) )
        
        except ModelAssertionViolation:
            
            return -numpy.inf
        
        except:
            
            raise
        
        if not numpy.isfinite( logLike ):
            
            return -numpy.inf
        
        else:
            
            return logLike
    
    def posterior( self, trialValues ):
        
        #Here we don't use the self._logp nor the self._logLike to
        #avoid looping twice over the parameters (for speed)
        
        #Assign this trial values to the parameters and
        #store the corresponding values for the priors
        
        lps = []
        
        for i,( srcName, paramName ) in enumerate( self.freeParameters.keys() ):
            
            thisParam = self.likelihoodModel.parameters[ srcName ][ paramName ]
                  
            thisParam.setValue( trialValues[i] )
            
            pval = thisParam.getPriorValue()
            
            if not numpy.isfinite( pval ):
                
                return -numpy.inf
            
            lps.append( pval )
            
        logLike = numpy.sum( map( lambda dataset: dataset.getLogLike(), self.dataList.values() ) )
        
        if not numpy.isfinite( logLike ):
            
            return -numpy.inf
        
        logPrior = numpy.sum( lps )
        
        return logLike + logPrior
    
    def _getStartingPoint( self, nwalkers, variance = 0.1 ):
        
        #Generate the starting points for the walkers by getting random
        #values for the parameters close to the current value
        
        #Fractional variance for randomization
        #(0.1 means var = 0.1 * value )
                
        p0 = [ ]
        
        for i in range( nwalkers ):
            
            thisP0 = []
            
            for (srcName, paramName) in self.freeParameters.keys():
                        
                thisPar = self.likelihoodModel.parameters[ srcName ][ paramName ]
                
                thisVal = thisPar.getRandomizedValue( variance )
                
                thisP0.append( thisVal )
            
            p0.append( thisP0 )
        
        return p0     
        
    
    def samplePT( self, ntemps, nwalkers, burn_in, nsamples ):
        '''
        Sample with parallel tempering
        '''
        
        self.freeParameters = self.likelihoodModel.getFreeParameters()
                
        ndim = len( self.freeParameters.keys() )
        
        sampler = emcee.PTSampler( ntemps, nwalkers, ndim, self._logLike, self._logp )
        
        #Get one starting point for each temperature
        
        p0 = numpy.empty( ( ntemps, nwalkers, ndim ) )
        
        for i in range( ntemps ):
            
            p0[i,:,:] = self._getStartingPoint( nwalkers )
        
        print("Running burn-in of %s samples...\n" % burn_in )
        
        p, lnprob, lnlike = sampleWithProgress( p0, sampler, burn_in )
        
        #Reset sampler
        
        sampler.reset()
        
        print("\nSampling...\n")
        
        p, lnprob, lnlike = sampleWithProgress( p, sampler, nsamples, 
                                                lnprob0 = lnprob, lnlike0 = lnlike )
        
        self.sampler = sampler
        self.samples = sampler.flatchain.reshape(-1, sampler.flatchain.shape[-1])
        
        return self.getSamples()
        
    
    def sample( self, nwalkers, burn_in, nsamples ):
        
        self.freeParameters = self.likelihoodModel.getFreeParameters()
                
        ndim = len( self.freeParameters.keys() )
        
        #Get starting point
        
        p0 = self._getStartingPoint( nwalkers )
        
        if threeML_config['parallel']['use-parallel']:
        
            c = ParallelClient()
            view = c[:]
        
            sampler = emcee.EnsembleSampler( nwalkers, ndim, 
                                             self.posterior,
                                             pool = view) 
        
        else:
            
            sampler = emcee.EnsembleSampler( nwalkers, ndim, 
                                             self.posterior )
        
        print("Running burn-in of %s samples...\n" % burn_in )
        
        pos, prob, state = sampleWithProgress( p0, sampler, burn_in )
        
        #Reset sampler
        
        sampler.reset()
        
        #Run the true sampling
        
        print("\nSampling...\n")
                
        _ = sampleWithProgress( pos, sampler, nsamples, rstate0=state ) 
        
        #sampler.run_mcmc( pos, nsamples, rstate0=state )
        
        acc = numpy.mean( sampler.acceptance_fraction )
        
        print( "Mean acceptance fraction: %s" % acc )
        
        self.sampler = sampler
        self.samples = sampler.flatchain
        
        return self.getSamples()
    
    def getSamples( self ):
        
        if hasattr( self, "samples" ):
            
            return self.samples
        
        else:
            
            raise RuntimeError("You have to run the sampler first, using the sample() method")
    
    def cornerPlot( self, **kwargs ):
        
        if hasattr( self, "samples" ):
            
            labels = []
            priors = []
            
            for i,( srcName, paramName ) in enumerate( self.freeParameters.keys() ):
                
                thisLabel = "%s of %s" % ( paramName, srcName )
                
                labels.append( thisLabel )
                
                priors.append( self.likelihoodModel.parameters[ srcName ][ paramName ].prior )
            
            fig = corner( self.samples, labels=labels, 
                          quantiles=[0.16, 0.50, 0.84],
                          priors = priors, **kwargs )
            
            return fig
        
        else:
            
            raise RuntimeError("You have to run the sampler first, using the sample() method")



#Author: L. Tibaldo, ltibaldo@slac.stanford.edu
from threeML.models.spatialmodel import SpatialModel
from threeML.models.Parameter import Parameter, SpatialParameter
import numpy as np
from angsep import angsep
import collections


class Ring(SpatialModel):
    
    def setup(self):
        self.functionName        = "Ring"
        self.formula             = r'$f({\rm RA, Dec}) = \left(\frac{180^\circ}{\pi}\right)^2 \frac{1}{\pi r^2 (1-f^2)} \, \left({\rm angsep} ({\rm RA, Dec, RA_0, Dec_0}) < r \right) \left({\rm angsep} ({\rm RA, Dec, RA_0, Dec_0}) > f \cdot r \right)$'
        self.parameters          = collections.OrderedDict()
        self.parameters['RA0']     = Parameter('RA0',1.,0,360,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['Dec0']     = Parameter('Dec0',1.,-90,90,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['radius'] = SpatialParameter('radius',0.1,0,20,0.01,fixed=False,nuisance=False,dataset=None)
        self.parameters['inner_frac'] = SpatialParameter('inner_frac',0.1,0,1,0.01,fixed=False,nuisance=False,dataset=None)
        
        self.ncalls              = 0
    
    
    def __call__(self,RA,Dec,energy):
        self.ncalls             += 1
        RA0                         = self.parameters['RA0'].value
        Dec0                        = self.parameters['Dec0'].value
        radius                      = self.parameters['radius'].getValue(energy)
        inner_frac                  = self.parameters['inner_frac'].getValue(energy)
        
        return np.power(180/np.pi,2)*1./(np.pi*radius**2*(1-inner_frac**2)) * (angsep(RA,Dec,RA0,Dec0)<radius)*(angsep(RA,Dec,RA0,Dec0)>(inner_frac*radius))

    def integratedFlux(self,energy):
    
        return 1.

class EllipticalRing(SpatialModel):
    
    def setup(self):
        self.functionName        = "EllipticalRing"
        self.formula             = r'$f({\rm RA, Dec}) = \left(\frac{180^\circ}{\pi}\right)^2 \frac{1}{\pi a^2 (1-f^2) \sqrt{1-e^2}} \,\left[ \left(\frac{\cos\theta \,{\rm RA} + \sin\theta\, {\rm Dec} - {\rm RA_0}}{a^2} + \frac{-\sin\theta\, {\rm RA} + \cos \theta \,{\rm Dec}- {\rm Dec_0}}{a^2 (1-e^2)}\right) < 1 \right] \left[ \left(\frac{\cos\theta \,{\rm RA} + \sin\theta\, {\rm Dec} - {\rm RA_0}}{a^2} + \frac{-\sin\theta\, {\rm RA} + \cos \theta \,{\rm Dec}- {\rm Dec_0}}{a^2 (1-e^2)}\right) >f^2 \right]$'
        self.parameters          = collections.OrderedDict()
        self.parameters['RA0']     = Parameter('RA0',1.,0,360,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['Dec0']     = Parameter('Dec0',1.,-90,90,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['maj_saxis'] = SpatialParameter('maj_saxis',0.1,0,10,0.01,fixed=False,nuisance=False,dataset=None)
        self.parameters['eccentricity'] = SpatialParameter('eccentricity',0.7,0,1,0.01,fixed=False,nuisance=False,dataset=None)
        self.parameters['angle'] = SpatialParameter('angle',0.,0,180,1.,fixed=False,nuisance=False,dataset=None)
        self.parameters['inner_frac'] = SpatialParameter('inner_frac',0.1,0,1,0.01,fixed=False,nuisance=False,dataset=None)
        
        self.ncalls              = 0
    
    
    def __call__(self,RA,Dec,energy):
        self.ncalls             += 1
        RA0                         = self.parameters['RA0'].value
        Dec0                        = self.parameters['Dec0'].value
        maj_saxis                   = self.parameters['maj_saxis'].getValue(energy)
        eccentricity                = self.parameters['eccentricity'].getValue(energy)
        angle                       = self.parameters['angle'].getValue(energy)
        inner_frac                  = self.parameters['inner_frac'].getValue(energy)
        
        
        s=np.sin(np.deg2rad(angle))
        c=np.cos(np.deg2rad(angle))
        return np.power(180/np.pi,2)*1./(np.pi*maj_saxis**2*(1-inner_frac**2)*np.sqrt(1-eccentricity**2))*(np.power(c*RA+s*Dec-c*RA0-s*Dec0,2)/maj_saxis**2+np.power(-s*RA+c*Dec+s*RA0-c*Dec0,2)/(maj_saxis**2*(1-eccentricity**2))<1)*(np.power(c*RA+s*Dec-c*RA0-s*Dec0,2)/maj_saxis**2+np.power(-s*RA+c*Dec+s*RA0-c*Dec0,2)/(maj_saxis**2*(1-eccentricity**2))>1*inner_frac**2
                                                                                                                                                                                                                                        )
    

    def integratedFlux(self,energy):
    
        return 1.



import warnings

from astropy.vo.client.vos_catalog import VOSCatalog
from astropy.vo.client import conesearch
from astropy.vo.client.exceptions import VOSError


class VirtualObservatoryCatalog(object):
    
    def __init__(self, name, url, description):
                
        self.catalog = VOSCatalog.create(name, url, description=description)
    
    def query(self, skycoord, radius):     
        
        with warnings.catch_warnings():
            
            #Ignore all warnings, which are many from the conesearch module
            
            warnings.simplefilter('ignore')
            
            try:
                
                votable = conesearch.conesearch(skycoord, radius, 
                                                catalog_db=self.catalog,
                                                verb=3, verbose=True,
                                                cache=False)
            
            except VOSError as exc:
                
                print(exc.message)
                return None
        
        out = self.applyFormat(votable)
        
        #This is needed to avoid strange errors
        del votable
        
        return out
    
    def applyFormat(self):
        
        raise NotImplementedError("You have to re-implement this!")



import collections
from Parameter import Parameter, SpatialParameter
from IPython.display import display, Latex, HTML
import abc



class SpatialModel(object):
    __metaclass__           = abc.ABCMeta # Make the user code certain
    

    def __init__(self):

        import collections        
        self.setup()








'''
Define the interface for a plugin class.
'''

import abc


class PluginPrototype(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def setModel(self, LikelihoodModelInstance):
        '''
        Set the model to be used in the joint minimization. Must be a LikelihoodModel instance.
        '''
        pass

    pass

    @abc.abstractmethod
    def getName(self):
        '''
        Return a name for this dataset (likely set during the constructor)
        '''
        pass

    pass

    @abc.abstractmethod
    def getLogLike(self):
        '''
        Return the value of the log-likelihood with the current values for the
        parameters
        '''
        pass

    pass

    @abc.abstractmethod
    def getNuisanceParameters(self):
        '''
        Return a list of nuisance parameters. Return an empty list if there
        are no nuisance parameters
        '''
        pass

    pass

    @abc.abstractmethod
    def innerFit(self):
        '''
        This is used for the profile likelihood. Keeping fixed all parameters in the
        LikelihoodModel, this method minimize the logLike over the remaining nuisance
        parameters, i.e., the parameters belonging only to the model for this
        particular detector. If there are no nuisance parameters, simply return the
        logLike value.
        '''
        pass

    pass


pass



from threeML.models.spectralmodel import SpectralModel
from threeML.models.Parameter import Parameter
import math
import scipy.integrate
import operator

import collections



class SmoothlyBrokenPowerlaw(SpectralModel):

    def setup(self):
        self.functionName        = "SmoothlyBrokenPowerlaw"
        self.formula             = r'\begin{equation}f(E) = A E^{\gamma}\end{equation}'

        self.parameters          = collections.OrderedDict()
        self.parameters['indx1'] = Parameter('indx1',-2.0,-10,10,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['indx2'] = Parameter('indx2',-2.0,-10,10,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['A']     = Parameter('A',1.0,1e-10,1e10,0.02,fixed=False,nuisance=False,dataset=None,normalization=True)
        self.parameters['breakEnergy']  = Parameter('breakEnergy',300.,1.,1.e8,1,fixed=False)
        self.parameters['breakScale']  = Parameter('breakScale',1.,1.e-2,1.e1,.1,fixed=False)
        self.parameters['Epiv']  = Parameter('Epiv',300.,1.e0,1.e6,1.,fixed=True)
        
    
        self.ncalls              = 0
    
        #def integral(e1,e2):
        #    a                      = self.parameters['gamma'].value
        #    piv                    = self.parameters['Epiv'].value
        #    norm                   = self.parameters['A'].value
#      
#            if(a!=-1):
#                def f(energy):
#                    return self.parameters['A'].value * math.pow(energy/piv,a+1)/(a+1)
#            else:
#                def f(energy):
#                    return self.parameters['A'].value * math.pow(energy/piv)
#            return f(e2)-f(e1)
        self.integral            = None #Check this!
 
  
    def __call__(self,ene):
        self.ncalls             += 1
        pivot                      = self.parameters['Epiv'].value
        norm                     = self.parameters['A'].value
        indx1                    = self.parameters['indx1'].value
        indx2                    = self.parameters['indx2'].value
        breakE                   = self.parameters['breakEnergy'].value
        breakScale               = self.parameters['breakScale'].value

        val = numpy.zeros(ene.flatten().shape[0])
            
        

        B = (indx1 + indx2)/2.0
        M = (indx2 - indx1)/2.0

        arg_piv = numpy.log10(pivot/breakE)/breakScale

        if arg_piv < -6.0:

            pcosh_piv = M * breakScale * (-arg_piv-numpy.log(2.0))

        elif arg_piv > 4.0:

            pcosh_piv = M * breakScale * (arg_piv - numpy.log(2.0))

        else:

            pcosh_piv = M * breakScale * (numpy.log( (numpy.exp(arg_piv) + numpy.exp(-arg_piv))/2.0 ))



        arg = log10(ene/breakE)/breakScale


        idx1 =  arg < -6.0
        idx2 =  arg >  4.0
        idx3 =  ~numpy.logical_or(idx1,idx2)

        pcosh = numpy.zeros(ene.flatten().shape[0])

        pcosh[idx1] = M * breakScale * (-arg[idx1]-numpy.log(2.0))

        pcosh[idx2] = M * breakScale * (arg[idx2] - numpy.log(2.0))

        pcosh[idx3] = M * breakScale * (numpy.log( (numpy.exp(arg[idx3]) + numpy.exp(-arg[idx3]))/2.0 ))

        val = norm * numpy.power(ene/pivot,B)*numpy.power(10.,pcosh-pcosh_piv)

        return val

   
  
    def photonFlux(self,e1,e2):
        return self.integral(e1,e2)
  
    def energyFlux(self,e1,e2):
        pass
   
    
        #return (eF(e2)-eF(e1))*keVtoErg




from threeML.models.spectralmodel import SpectralModel
from threeML.models.Parameter import Parameter
import numpy
import math
import scipy.integrate
import operator

import collections


class LogParabola(SpectralModel):
    def setup(self):
            self.functionName        = "LogParabola"
            self.formula             = r'\begin{equation}f(E) = A E^{\gamma+\beta \log(E)}\end{equation}'
            self.parameters          = collections.OrderedDict()
            self.parameters['gamma'] = Parameter('gamma',-1.5,-10,10,0.1,fixed=False,nuisance=False,dataset=None)
            self.parameters['beta'] = Parameter('beta',-0.5,-10,10,0.1,fixed=False,nuisance=False,dataset=None)
            self.parameters['logA']     = Parameter('logA',-10,-40,20,1,fixed=False,nuisance=False,dataset=None,normalization=True)
            self.parameters['Epiv']  = Parameter('Epiv',1.0,1e-10,1e10,1,fixed=True)
    
            self.ncalls              = 0
    


            def integral(e1,e2):
                return self((e1+e2)/2.0)*(e2-e1)
            self.integral            = integral
    
    def __call__(self,e):
          
          self.ncalls             += 1
          piv                     = self.parameters['Epiv'].value
          gamma                = self.parameters['gamma'].value
          beta                    = self.parameters['beta'].value
          norm                     = pow(10, self.parameters['logA'].value)
          
          energies = numpy.array( e, ndmin=1, copy=False)          
 
          return norm * (energies/piv)**(gamma+beta*numpy.log10(energies/piv))
  
  

    def photonFlux(self,e1,e2):
        return self.integral(e1,e2)
  
  #def energyFlux(self,e1,e2):
  #  a                        = self.parameters['gamma'].value
  #  piv                      = self.parameters['Epiv'].value
  #  if(a!=-2):
  #    def eF(e):
  #      return numpy.maximum(self.parameters['A'].value * numpy.power(e/piv,2-a)/(2-a),1e-30)
  #  else:
  #    def eF(e):
  #      return numpy.maximum(self.parameters['A'].value * numpy.log(e/piv),1e-30)
  #  pass
    
   # return (eF(e2)-eF(e1))*keVtoErg
pass


class LogParabolaEp(SpectralModel):
    def setup(self):
            self.functionName        = "LogParabolaEp"
            self.formula             = r'\begin{equation}f(E) = \frac{S_{p}}{E^{2}} (E/E_{p})^{-b \log{E/E_{p}}}\end{equation}'
            self.parameters          = collections.OrderedDict()
            self.parameters['b'] = Parameter('b',0.6,0,5,0.1,fixed=False,nuisance=False,dataset=None)
            self.parameters['Sp'] = Parameter('Sp',1,1e-5,1e5,0.1,normalization=True, fixed=False,nuisance=False,dataset=None)
            self.parameters['Ep']  = Parameter('Ep',300.0,1.0,1e6,100,fixed=False,nuisance=False,dataset=None)

            def integral(e1,e2):
                return self((e1+e2)/2.0)*(e2-e1)
            
            self.integral            = integral
    
    def __call__(self, e ):
          
          energies = numpy.array( e, ndmin=1, copy=False)
          
          b = self.parameters['b'].value
          Sp = self.parameters['Sp'].value
          Ep = self.parameters['Ep'].value
          
          eep = energies / Ep
          
          out = Sp / numpy.power( energies, 2 ) * numpy.power( eep, -b * numpy.log( eep ) )
          
          if(out.shape[0]==1):
          
            return out[0]
          
          else:
          
            return out  



from threeML.models.spectralmodel import SpectralModel
from threeML.models.Parameter import Parameter
import math
import scipy.integrate
import operator
import numpy


import collections



class ExponentialCutoff(SpectralModel):

    def setup(self):
        self.functionName        = "ExponentialCutoff"
        self.formula             = r'$f(E) = A {\rm exp}\left(-E/E_{\rm fold}   \right)$'
        self.parameters          = collections.OrderedDict()
        self.parameters['A']     = Parameter('A',1.,1.E-10,1.E10,0.1,fixed=False,nuisance=False,dataset=None,normalization=True)
        self.parameters['logEfold'] = Parameter('logEfold',2,0,20,0.1,fixed=False,nuisance=False,dataset=None)
            
        self.ncalls              = 0
    
        def integral(e1,e2):
            eFold                      = pow(10,self.parameters['logEfold'].value)
            A                          = self.parameters['A'].value

            
            def f(x):

                return -A * eFold * numpy.exp(-x/eFold)

            return f(e2) - f(e1)
            
        self.integral            = integral
 
  
    def __call__(self,e):
        self.ncalls             += 1
        eFold                      = pow(10, self.parameters['logEfold'].value)
        A                          = self.parameters['A'].value
        
        energies = numpy.array( e, ndmin=1, copy=False)
        
        return numpy.maximum( numpy.exp(-energies/eFold), 1e-30)
   
  
    def photonFlux(self,e1,e2):
        return self.integral(e1,e2)
  
    def energyFlux(self,e1,e2):
        eFold                      = pow(10,self.parameters['logEfold'].value)
        A                          = self.parameters['A'].value

        def eF(x):

            return -numpy.exp(-x/eFold)*(x*eFold + (eFold*eFold ))

        return (eF(e2) - eF(e1))*self.keVtoErg

        





import os
import sys
import glob
import inspect
import imp


# This dynamically loads a module and return it in a variable

def import_module(module_name):

    # Fast path: see if the module has already been imported.

    try:

        return sys.modules[module_name]

    except KeyError:

        pass

    # If any of the following calls raises an exception,
    # there's a problem we can't handle -- let the caller handle it.

    fp, pathname, description = imp.find_module(module_name)

    try:

        return imp.load_module(module_name, fp, pathname, description)

    except:

        raise

    finally:

        # Since we may exit via an exception, close fp explicitly.

        if fp:

            fp.close()

# Import in the current namespace everything under the
# models directory

# This must be here before the automatic import of subpackages,
# otherwise we will incur in weird issues with other packages
# using similar names (for example, the io package)

from .models.PointSource import PointSource
from .models.ExtendedSource import ExtendedSource
from .models.LikelihoodModel import LikelihoodModel
from .models.spectralmodel import *
from .exceptions import CustomExceptions
from .exceptions.CustomExceptions import custom_warnings
from .plugin_prototype import PluginPrototype

# Import the builtinModels

from .models.fluxModels import *

try:

    from cthreeML.pyModelInterface import pyToCppModelInterface

except ImportError:

    custom_warnings.warn("cthreeML is not installed. You will not be able to use plugins which require it.",
                         CustomExceptions.CppInterfaceNotAvailable)

# Import the classic Maximum Likelihood Estimation package

from .classicMLE.jointLikelihood import JointLikelihood

# Import the DataList class

from dataList import DataList


# Find the directory containing 3ML

threeML_dir = os.path.abspath(os.path.dirname(__file__))

# Import all modules here

sys.path.insert(0, threeML_dir)
mods = [os.path.basename(f)[:-3] for f in glob.glob(os.path.join(threeML_dir, "*.py"))]

# Filter out __init__

modsToImport = filter(lambda x: x.find("__init__") < 0, mods)

# Import everything in current directory

for mod in modsToImport:
    exec ("from %s import *" % mod)

# Now look for plugins

plugins_dir = os.path.join(os.path.dirname(__file__), "plugins")
sys.path.insert(1, plugins_dir)
found_plugins = glob.glob(os.path.join(plugins_dir, "*.py"))

# Filter out __init__

found_plugins = filter(lambda x: x.find("__init__") < 0, found_plugins)

msgs = []

for i, plug in enumerate(found_plugins):

    # Loop over each candidates plugins

    try:

        thisPlugin = import_module(os.path.basename(".".join(plug.split(".")[:-1])))

    except ImportError:

        custom_warnings.warn("Could not import plugin %s. Do you have the relative instrument software installed "
                             "and configured?" % plug, CustomExceptions.CannotImportPlugin)
        continue

    # Get the classes within this module

    classes = inspect.getmembers(thisPlugin, lambda x: inspect.isclass(x) and inspect.getmodule(x) == thisPlugin)

    for name, cls in classes:

        if not issubclass(cls, PluginPrototype):

            # This is not a plugin

            pass

        else:

            string = "%s for %s" % (cls.__name__, thisPlugin.__instrument_name)
            msgs.append("* %-60s available" % string)

            # import it again in the uppermost namespace

            exec ("from %s import %s" % (os.path.basename(thisPlugin.__name__), cls.__name__))


def get_available_plugins():

    print("Plugins:\n")

    print("\n".join(msgs))

from .parallel.ParallelClient import parallel_computation

# Now read the configuration and make it available as threeML_config
from .config.Config import threeML_config



#author L. Tibaldo (ltibaldo@slac.stanford.edu)

from threeML.models.spatialmodel import SpatialModel
from threeML.models.Parameter import Parameter, SpatialParameter
import pyfits as pf
import numpy as np
from astropy.wcs import WCS
from astropy import coordinates as coord
from astropy import units as u

def return_lonlat(RA,Dec,w):
        #if the map is in Galactic coords need to convert from Celestial
        if w.wcs.lngtyp=='GLON':
            c = coord.ICRS(ra=RA, dec=Dec, unit=(u.degree, u.degree))
            lon = c.galactic.l
            lat = c.galactic.b
        else:
            lon, lat = RA, Dec

        return lon, lat


def fp_linear_interp(pos,map,filename):
        #4-point linear interpolation to determine the values at the requested coords
        px_0=pos[-1].astype('int')
        py_0=pos[-2].astype('int')
        px_arr=np.array([px_0,px_0,px_0+1,px_0+1])
        py_arr=np.array([py_0,py_0+1,py_0,py_0+1])
        b=pos[-1]-px_0
        a=1.-b
        d=pos[-2]-py_0
        c=1.-d
        if len(pos)==2:
            fp_pos=[py_arr,px_arr]
        else:
            fp_pos=[4*[pos[-3],py_arr,px_arr]]
        try:
            vals = map[fp_pos]
        except ValueError:
            print "The WCS map in file {} is not defined at the sky coordinates requested by the user".format(filename)
        vals=vals*np.array([a*c,a*d,b*c,b*d])
        vals=np.sum(vals,axis=0)

        return vals


class WCSSpatialMap(SpatialModel):

    def setup(self,file):
        #assumes that file is a fits file with WCS map in hdu 0
        self.filename=file
        self.w=WCS(file)
        self.map=np.nan_to_num(pf.getdata(file,0))
        
        self.functionName        = "WCSSpatialMap"
        self.parameters          = collections.OrderedDict()
        self.parameters['Norm']  = Parameter('Norm',1.,0.,1.e5,0.1,fixed=True,nuisance=False,dataset=None)
        self.ncalls              = 0

    def __call__(self,RA,Dec):
        #you should make sure to oversample the original map when you call this function
        self.ncalls += 1
        Norm = self.parameters['Norm'].value
        #if the map is in Galactic coords need to convert from Celestial
        lon, lat = return_lonlat(RA,Dec,self.w)
        px,py = self.w.wcs_world2pix(lon, lat, 1)
        pos=[py,px]
        vals=fp_linear_interp(vec,self.map,self.filename)

        return Norm*vals


class WCSMapCube(SpatialModel):

    def setup(self,file):
        #assumes that file is a fits file with WCS map in hdu 0
        self.filename=file
        self.w=WCS(file)
        self.cube=np.nan_to_num(pf.getdata(file,0))
        self.energies=pf.getdata(file,1)['Energy']
        
        self.functionName        = "WCSMapCube"
        self.parameters          = collections.OrderedDict()
        self.parameters['Norm']  = Parameter('Norm',1.,0.,1.e5,0.1,fixed=True,nuisance=False,dataset=None)
        self.ncalls              = 0

    def __call__(self,RA,Dec,energy):
        #you should make sure to oversample the original map when you call this function
        self.ncalls += 1
        Norm = self.parameters['Norm'].value
        #if the map is in Galactic coords need to convert from Celestial
        lon, lat = return_lonlat(RA,Dec,self.w)
        px,py = self.w.wcs_world2pix(lon, lat, 0, 1)
        #determine values at requested energies by PL interp on the nearest available in the model
        E_0=np.max(np.where(energies<energy[:,np.newaxis]),axis=1)
        pos_0=[E_0,py,px]
        vals_0=fp_linear_interp(pos_0,self.cube,self.filename)
        pos_1=[E_0+1,py,px]
        vals_1=fp_linear_interp(pos_1,self.cube,self.filename)
        logE0=np.log(self.energies[E_0:np.newaxis,np.newaxis])
        logE1=np.log(self.energies[E_0+1:np.newaxis,np.newaxis])
        gamma=-(np.log(vals_1)-np.log(vals_0))/(logE1-logE0)
        vals=vals_0-gamma*(np.log(energy)-logE0)
        vals=np.exp(vals)

        return Norm*vals



import UnbinnedAnalysis
import BinnedAnalysis
import pyLikelihood as pyLike
import os
import shutil
import numpy
import astropy.io.fits as pyfits
from threeML.plugin_prototype import PluginPrototype
from threeML.models.Parameter import Parameter

from GtBurst import LikelihoodComponent
from GtBurst import IRFS

from GtBurst import FuncFactory

import matplotlib.pyplot as plt

from matplotlib import gridspec

__instrument_name = "Fermi LAT (standard classes)"

class myPointSource(LikelihoodComponent.GenericSource):
  def __init__( self, source , name ):
    
    self.source                        = source
    self.source.name                   = name
    
    

class LikelihoodModelConverter(object):
  
  def __init__(self, likelihoodModel, irfs):
    
    self.likelihoodModel      = likelihoodModel
    
    self.irfs                 = irfs
  
  def setFileSpectrumEnergies(self, emin_kev, emax_kev, nEnergies ):
    
    self.energiesKeV          = numpy.logspace( numpy.log10(emin_kev), 
                                                numpy.log10(emax_kev),
                                                nEnergies )
    
  
  def writeXml(self, xmlfile, ra, dec, roi):
    
    #Loop through all the sources in the likelihood model and generate a FileSpectrum
    #for all of them. This is necessary to allow the FermiLATLike class
    #to update the spectrum in pyLikelihood without having to write and read a .xml file
    #on the disk
    
    allSourcesForPyLike       = []
    
    nPtsrc                    = self.likelihoodModel.getNumberOfPointSources()
    
    for ip in range(nPtsrc):
      
      allSourcesForPyLike.append( self._makeFileSpectrum( ip ) )
    
    #Now the same for extended sources
    
    nExtSrc                   = self.likelihoodModel.getNumberOfExtendedSources()
    
    if( nExtSrc > 0 ):
      
      raise NotImplemented("Cannot support extended sources yet!")
    
    iso                       = LikelihoodComponent.IsotropicTemplate( self.irfs )
        
    iso.source.spectrum.Normalization.max = 1.5
    iso.source.spectrum.Normalization.min = 0.5
    iso.source.spectrum.setAttributes()
    
    allSourcesForPyLike.append( iso )
    
    gal                       = LikelihoodComponent.GalaxyAndExtragalacticDiffuse( 
                                                 self.irfs, ra, dec, 2.5*roi )
    gal.source.spectrum.Value.max = 1.5
    gal.source.spectrum.Value.min = 0.5
    gal.source.spectrum.setAttributes()
    
    allSourcesForPyLike.append( gal )
    
    #Now generate the xml file with also the Galactic and Isotropic diffuse
    #templates
    xml                          = LikelihoodComponent.LikelihoodModel()
    xml.addSources(*allSourcesForPyLike)
    xml.writeXML(xmlfile)
  
  def _makeFileSpectrum(self, ip):
    
    name                      = self.likelihoodModel.getPointSourceName( ip )
        
    values                    = self.likelihoodModel.getPointSourceFluxes( ip , 
                                                               self.energiesKeV )
    
    tempName                  = "__%s.txt" % name
    
    with open( tempName , "w+") as f:
      
      for e,v in zip( self.energiesKeV, values ):
        
        #Gtlike needs energies in MeV and fluxes in ph/MeV/cm2)
        
        f.write( "%s %s\n" % ( e / 1000.0 , v * 1000.0 ) )
    
    #p                         = fileFunction.parameter("Normalization")
    #p.setBounds(1-float(effAreaAllowedSize),1+effAreaAllowedSize)
    
    #Now generate the XML source wrapper
    #This is convoluted, but that's the ST way of doing things...
    #The final product is a class with a writeXml method
        
    src                        = '\n'.join( (('<source name= "%s" ' % name) + 'type="PointSource">',
                                           '   <spectrum type="PowerLaw2"/>',
                                           '   <!-- point source units are cm^-2 s^-1 MeV^-1 -->',
                                           '   <spatialModel type="SkyDirFunction"/>',
                                           '</source>\n') )
    src                        = FuncFactory.minidom.parseString(src).getElementsByTagName('source')[0]
    src                        = FuncFactory.Source(src)
        
    src.spectrum               = FuncFactory.FileFunction()
    src.spectrum.file          = tempName
    src.spectrum.parameters['Normalization'].value = 1.0
    src.spectrum.parameters['Normalization'].max = 1.1
    src.spectrum.parameters['Normalization'].min = 0.9
    src.spectrum.setAttributes()
    src.deleteChildElements('spectrum')
    src.node.appendChild(src.spectrum.node)
    
    src.spatialModel           = FuncFactory.SkyDirFunction()
    src.deleteChildElements('spatialModel')
    src.node.appendChild(src.spatialModel.node)
    
    ra,dec                     = self.likelihoodModel.getPointSourcePosition( ip )
    
    src.spatialModel.RA.value  = ra
    src.spatialModel.DEC.value = dec
    src.spatialModel.setAttributes()
    src.setAttributes()
    
    return myPointSource( src, name )


class FermiLATLike(PluginPrototype):
  
  def __init__(self, name, eventFile, ft2File, livetimeCube, kind, *args):
    
    self.name                 = name
    
    #Read the ROI cut
    cc                           = pyLike.RoiCuts()
    cc.readCuts(eventFile, "EVENTS")
    self.ra, self.dec, self.rad  = cc.roiCone()

    #Read the IRF selection
    c                         = pyLike.Cuts(eventFile, "EVENTS")
    self.irf                  = c.CALDB_implied_irfs()
    
    self.ft2File              = ft2File   
    self.livetimeCube         = livetimeCube
    
    #These are the boundaries and the number of energies for the computation
    #of the model (in keV)
    self.emin                 = 1e4
    self.emax                 = 5e8
    self.Nenergies            = 1200
    
    #This is the limit on the effective area correction factor,
    #which is a multiplicative factor in front of the whole model
    #to account for inter-calibration issues. By default it can vary
    #by 10%. This can be changed by issuing:
    #FermiLATUnbinnedLikeInstance.effCorrLimit = [new limit]
    #where for example a [new limit] of 0.2 allow for an effective
    #area correction up to +/- 20 %
    
    self.effCorrLimit         = 0.1
    
    if( kind.upper() != "UNBINNED" and kind.upper() != "BINNED"):
      
      raise RuntimeError("Accepted values for the kind parameter are: " +
                         "binned, unbinned. You specified: %s" %(kind))
    
    else:
      
      self.kind               = kind.upper()
    
        
    if(kind.upper()=='UNBINNED'):
      exposureMap             = args[0]
      self.eventFile          = eventFile
      self.exposureMap          = exposureMap
      #Read the files and generate the pyLikelihood object
      self.obs                  = UnbinnedAnalysis.UnbinnedObs(self.eventFile,
                                             self.ft2File,
                                             expMap=self.exposureMap,
                                             expCube=self.livetimeCube,
                                             irfs=self.irf)
    
    elif(kind.upper()=="BINNED"):
       sourceMaps,binnedExpoMap  = args
       self.sourceMaps           = sourceMaps
       self.binnedExpoMap        = binnedExpoMap
       
       self.obs                  = BinnedAnalysis.BinnedObs(srcMaps=self.sourceMaps,
                                 	     expCube=self.livetimeCube,
                                 	     binnedExpMap=self.binnedExpoMap,
                                 	     irfs=self.irf)
    pass
        
    #Activate inner minimization by default
    self.setInnerMinimization( True )
    
  pass
    
  def setModel(self, likelihoodModel):
    '''
    Set the model to be used in the joint minimization. 
    Must be a LikelihoodModel instance.
    '''
    self.lmc                  = LikelihoodModelConverter( likelihoodModel, 
                                                          self.irf )
    
    self.lmc.setFileSpectrumEnergies( self.emin, self.emax, self.Nenergies )
    
    xmlFile                    = '__xmlFile.xml'
    
    self.lmc.writeXml( xmlFile, self.ra, self.dec, self.rad )
    
    if( self.kind == "BINNED" ):
      self.like                  = BinnedAnalysis.BinnedAnalysis( self.obs,
                                                                xmlFile,
					                      optimizer='NEWMINUIT')

    else:
      
      self.like                  = UnbinnedAnalysis.UnbinnedAnalysis( self.obs,
                                                                xmlFile,
					                      optimizer='NEWMINUIT')
      
    
    self.likelihoodModel       = likelihoodModel
    
    #Here we need also to compute the logLike value, so that the model
    #in the XML file will be chanded if needed
    dumb                      = self.getLogLike()
    
    #Build the list of the nuisance parameters
    self._setNuisanceParameters()
  pass
  
  def getName(self):
    '''
    Return a name for this dataset (likely set during the constructor)
    '''
    return self.name
  pass
  
  def setInnerMinimization( self, s ):
    
    self.innerMinimization = bool( s )
  
  def innerFit(self):
    '''
    This is used for the profile likelihood. Keeping fixed all parameters in the
    modelManager, this method minimize the logLike over the remaining nuisance
    parameters, i.e., the parameters belonging only to the model for this
    particular detector
    '''
    self._updateGtlikeModel()
    
    if not self.innerMinimization:
        
        return self.like.logLike.value()
    
    try:
      #Use .optimize instead of .fit because we don't need the errors
      #(.optimize is faster than .fit)
      self.like.optimize(0)
    except:
      #This is necessary because sometimes fitting algorithms go and explore extreme region of the
      #parameter space, which might turn out to give strange model shapes and therefore
      #problems in the likelihood fit
      print("Warning: failed likelihood fit (probably parameters are too extreme).")
      return 1e5
    else:
      #Update the value for the nuisance parameters
      for par in self.nuisanceParameters.values():
        newValue             = self.getNuisanceParameterValue(par.name)
        par.setValue(newValue)
      pass
      
      return self.like.logLike.value()  
  pass
  
  #def _initFileFunction(self):
    #If the point source is already in the model, delete it
    #(like.deleteSource() returns the model)
    #gtlikeSrcModel            = self.like.deleteSource(self.modelManager.name)
    
    #Get the new model for the source (with the latest parameter values)
    #and add it back to the likelihood model
  #  self._getNewGtlikeModel(gtlikeSrcModel,self.effCorrLimit)
  #  self.like.addSource(gtlikeSrcModel)
    
    #Slow! But no other options at the moment
  #  self.like.writeXml(self.xmlModel)
  #  self.like.logLike.reReadXml(self.xmlModel)
  #pass
  
  #def _getNewGtlikeModel(self,currentGtlikeModel,effAreaAllowedSize=0.1):
    #Write on disk the current model
  #  tempName                  = os.path.join(
  #                              os.path.dirname(
  #                              os.path.abspath(self.xmlModel)
  #                                             ), 
  #                             '__fileSpectrum.txt')
  #  
  #  #This will recompute the model if necessary
  #  self.modelManager.writeToFile(tempName,self.emin,self.emax,self.Nenergies)
  #  
  #  #Generate a new FileFunction spectrum and assign it to the source
  #  fileFunction              = pyLike.FileFunction()
  #  fileFunction.readFunction(tempName)
  #  fileFunction.setParam("Normalization",1)
  #  p                         = fileFunction.parameter("Normalization")
  #  p.setBounds(1-float(effAreaAllowedSize),1+effAreaAllowedSize)
  #  
  #  currentGtlikeModel.setSpectrum(fileFunction)
  #pass  
  
  def _updateGtlikeModel(self):    
    '''
    #Slow! But no other options at the moment
    self.like.writeXml(self.xmlModel)
    self.like.logLike.reReadXml(self.xmlModel)
    '''
    
    energies                  = self.lmc.energiesKeV
    
    for srcName in self.likelihoodModel.pointSources.dict.keys():
      
      values                  = self.likelihoodModel.getPointSourceFluxes(
                                                         srcName,
                                                         energies
                                                         )      
      gtlikeSrcModel          = self.like[srcName]
    
      my_function             = gtlikeSrcModel.getSrcFuncs()['Spectrum']
      my_file_function        = pyLike.FileFunction_cast(my_function)
    
      my_file_function.setParam("Normalization",1)
      my_file_function.setSpectrum( energies / 1000.0, values * 1000.0 )
      gtlikeSrcModel.setSpectrum(my_file_function)
    
    #TODO: extended sources
    
  pass
  
  def getLogLike(self):
    '''
    Return the value of the log-likelihood with the current values for the
    parameters stored in the ModelManager instance
    '''
    self._updateGtlikeModel()
    try:
    
      value                   = self.like.logLike.value()
    
    except:
      raise
      
      value                   = 1e5
    
    pass
    
    return value
  pass
  
  def getModelAndData(self):
  
    fake = numpy.array([])
  
    return fake,fake,fake,fake
  
  pass
  
  def display( self ):
    
    e1 = self.like.energies[:-1]
    e2 = self.like.energies[1:]
    
    ec = ( e1 + e2 ) / 2.0
    de = ( e2 - e1 ) / 2.0
    
    sum_model = numpy.zeros_like( self.like._srcCnts( self.like.sourceNames()[0] ) )
    
    fig = plt.figure()
    
    gs = gridspec.GridSpec(2,1, height_ratios=[2,1])
    gs.update(hspace=0)
    
    sub = plt.subplot( gs[0] )
    
    for sourceName in self.like.sourceNames():
      
      sum_model = sum_model + self.like._srcCnts(sourceName)
      
      sub.plot( ec, self.like._srcCnts( sourceName ), label = sourceName )
   
    sub.plot( ec, sum_model, label = 'Total Model' )
   
    sub.errorbar( ec, self.like.nobs, xerr=de, 
                  yerr = numpy.sqrt( self.like.nobs ), 
                  fmt='.', label='Counts')
    
    plt.legend( bbox_to_anchor=(1.05, 1), loc=2, numpoints=1)
    
    #Residuals
    
    sub1 = plt.subplot( gs[1] )
    
    #Using model variance to account for low statistic
    
    resid = ( self.like.nobs - sum_model ) / sum_model
    
    sub1.axhline( 0, linestyle='--' )
    sub1.errorbar( ec, resid, xerr=de, 
                   yerr=numpy.sqrt(self.like.nobs) / sum_model,
                   capsize=0, fmt='.' )
    
    sub.set_xscale("log")
    sub.set_yscale("log", nonposy='clip')
    
    sub.set_ylabel("Counts per bin") 
    
    sub1.set_xscale("log")
    
    sub1.set_xlabel("Energy (MeV)")
    sub1.set_ylabel("(Mo. - data) / mo.")
    
    sub.set_xticks( [] )
    
    return fig
  
  def _setNuisanceParameters(self):

    #Get the list of the sources
    sources                   = list(self.like.model.srcNames)
    
    freeParamNames            = []
    for srcName in sources:
      thisNamesV              = pyLike.StringVector()
      thisSrc                 = self.like.logLike.getSource(srcName)
      thisSrc.spectrum().getFreeParamNames(thisNamesV)
      thisNames               = map(lambda x:"%s-%s" %(srcName,x),thisNamesV)
      freeParamNames.extend(thisNames)
    pass
    
    self.nuisanceParameters   = {}
    
    for name in freeParamNames:
      value                   = self.getNuisanceParameterValue(name)
      bounds                  = self.getNuisanceParameterBounds(name)
      delta                   = self.getNuisanceParameterDelta(name)
      self.nuisanceParameters[name] = Parameter(name,value,bounds[0],bounds[1],delta,nuisance=True)
    pass
    
  pass
  
  def getNuisanceParameters(self):
    '''
    Return a list of nuisance parameters. Return an empty list if there
    are no nuisance parameters
    '''
    return self.nuisanceParameters
  
  
  def getNuisanceParameterValue(self,paramName):
    src,pname                 = paramName.split("-")
    return self.like.model[src].funcs['Spectrum'].getParam(pname).getValue()
  pass
  
  def getNuisanceParameterBounds(self,paramName):
    src,pname                 = paramName.split("-")
    return list(self.like.model[src].funcs['Spectrum'].getParam(pname).getBounds())
  pass
  
  def getNuisanceParameterDelta(self,paramName):
    src,pname                 = paramName.split("-")
    value                     = self.like.model[src].funcs['Spectrum'].getParam(pname).getValue()
    return value/100.0
  pass
  
  def setNuisanceParameterValue(self,paramName,value):
    src,pname                 = paramName.split("-")
    self.like.model[src].funcs['Spectrum'].getParam(pname).setValue(value)
  pass
pass



# Author: L. Tibaldo, ltibaldo@slac.stanford.edu

from threeML.models.spatialmodel import SpatialModel
from threeML.models.Parameter import Parameter, SpatialParameter
import numpy as np
import collections
from angsep import angsep


class Disk(SpatialModel):
    def setup(self):
        self.functionName = "Disk"
        self.formula = r'$f({\rm RA, Dec}) = \left(\frac{180^\circ}{\pi}\right)^2 \frac{1}{\pi r^2} \, \left({\rm angsep} ({\rm RA, Dec, RA_0, Dec_0}) < r \right) $'
        self.parameters = collections.OrderedDict()
        self.parameters['RA0'] = Parameter('RA0', 1., 0, 360, 0.1, fixed=False, nuisance=False, dataset=None)
        self.parameters['Dec0'] = Parameter('Dec0', 1., -90, 90, 0.1, fixed=False, nuisance=False, dataset=None)
        self.parameters['radius'] = SpatialParameter('radius', 0.1, 0, 20, 0.01, fixed=False, nuisance=False,
                                                     dataset=None)

        self.ncalls = 0

    def __call__(self, RA, Dec, energy):
        self.ncalls += 1
        RA0 = self.parameters['RA0'].value
        Dec0 = self.parameters['Dec0'].value
        radius = self.parameters['radius'].getValue(energy)

        return np.power(180 / np.pi, 2) * 1. / (np.pi * radius ** 2) * (angsep(RA, Dec, RA0, Dec0) < radius)

    def integratedFlux(self, energy):
        return 1.

    def getBoundaries(self):

        maxRadius = self.parameters['radius'].maxValue

        minDec = max(-90.,self.parameters['Dec0'].value - maxRadius)
        maxDec = min(90.,self.parameters['Dec0'].value + maxRadius)

        maxAbsDec = max(np.absolute(minDec),np.absolute(maxDec))
        if maxAbsDec > 89. or maxRadius/np.cos(maxAbsDec*np.pi/180.) >= 180.:
            minRa = 0.
            maxRa = 360.
        else:
            minRa = self.parameters['RA0'].value - maxRadius/np.cos(maxAbsDec*np.pi/180.)
            maxRa = self.parameters['RA0'].value + maxRadius/np.cos(maxAbsDec*np.pi/180.)
            if minRa < 0.:
                minRa = minRa + 360.
            elif maxRa > 360.:
                maxRa = maxRa - 360.

        return minRa, maxRa, minDec, maxDec




#Author: L. Tibaldo, ltibaldo@slac.stanford.edu

from SourceModel import SourceModel
import numpy as np

import collections

class ExtendedSource(SourceModel):
    def __init__(self,name,spatialModel,spectralModel):

        self.name                 = name
        self.functionName         = "Extended source"
    
    
        if(not callable(spatialModel)):
        
            raise RuntimeError("The provided spatial model must be callable")
    
        else:
        
            self.spatialModel      = spatialModel


        if(not callable(spectralModel)):
    
            raise RuntimeError("The provided spectral model must be callable")
    
        else:
        
            self.spectralModel      = spectralModel

        pass
        
        self.parameters           = collections.OrderedDict()
        
        for k,v in self.spatialModel.parameters.iteritems():
            
            self.parameters[ k ] = v
        
        
        #Parameters from the spectral model are already imported elsewhere
        
        #for k,v in self.spectralModel.parameters.iteritems():
            
        #    self.parameters[ k ] = v
        
        self.formula = self.spatialModel.formula
        
    pass

    def getFlux(self,energies):
        return self.spatialModel.integratedFlux(energies)*self.spectralModel(energies)

    def getBrightness(self,RA,Dec,energies):
        return self.spatialModel(RA,Dec,energies)*self.spectralModel(energies)
    
    def getBoundaries(self):
        
        return self.spatialModel.getBoundaries()



pass



import astropy.io.fits as pyfits
import xml.etree.ElementTree as ET
import os

from threeML.plugins.gammaln import logfactorial
from threeML.io import fileUtils

import numpy
from threeML.plugins.ogip import OGIPPHA
from threeML.plugin_prototype import PluginPrototype
from threeML.models.Parameter import Parameter
from threeML.minimizer import minimization
from threeML.utils import Binner
import scipy.integrate

import warnings
import collections

import matplotlib.pyplot as plt
from matplotlib import gridspec


__instrument_name = "All OGIP-compliant instruments"

class GenericOGIPLike(PluginPrototype):
  
  def __init__(self, name, phafile, bkgfile, rspfile, arffile=None):
    
    self.name                 = name
    
    #Check that all file exists
    notExistant               = []
    
    inputFiles = [ phafile, bkgfile, rspfile ]
    
    for i in range( 3 ):
        
        #The file could contain a {#} specification, like spectrum.pha{3},
        #which indicate the 3rd spectrum in the spectrum.pha file
        
        inputFiles[i] = fileUtils.sanitizeFilename( inputFiles[i].split("{")[0] )
        
        if not fileUtils.fileExistingAndReadable( inputFiles[i] ):
            
            raise IOError("File %s does not exist or is not readable" % ( inputFiles[i] ) )
    
    phafile, bkgfile, rspfile = inputFiles
    
    #Check the arf, if provided
    if arffile is not None:
        
        arffile = fileUtils.sanitizeFilename( arffile.split("{")[0] )
        
        if not fileUtils.fileExistingAndReadable( arffile ):
            
            raise IOError("File %s does not exist or is not readable" % ( arf ) )
    
    self.phafile              = OGIPPHA(phafile,filetype='observed')
    self.exposure             = self.phafile.getExposure()
    self.bkgfile              = OGIPPHA(bkgfile,filetype="background")
    self.response             = Response( rspfile, arffile )    
    
    #Start with an empty mask (the user will overwrite it using the
    #setActiveMeasurement method)
    self.mask                 = numpy.asarray(
                                    numpy.ones(self.phafile.getRates().shape),
                                    numpy.bool)
    
    #Get the counts for this spectrum
    self.counts               = ( self.phafile.getRates()[self.mask]
                                  * self.exposure )
    
    #Check that counts is positive
    idx                       = (self.counts < 0)
    
    if(numpy.sum(idx) > 0):
      
      warnings.warn("The observed spectrum for %s " % self.name + 
                    "has negative channels! Fixing those to zero.", 
                     RuntimeWarning)
      self.counts[idx]        = 0
    
    pass
    
    #Get the background counts for this spectrum
    self.bkgCounts            = ( self.bkgfile.getRates()[self.mask]
                                  * self.exposure )
    
    
    #Check that bkgCounts is positive
    idx                       = (self.bkgCounts < 0)
    
    if(numpy.sum(idx) > 0):
      
      warnings.warn("The background spectrum for %s " % self.name + 
                    "has negative channels! Fixing those to zero.", 
                     RuntimeWarning)
      self.bkgCounts[idx]     = 0
    
    #Check that the observed counts are positive
    
    idx = self.counts < 0
    
    if numpy.sum( idx ) > 0:
       
       raise RuntimeError("Negative counts in observed spectrum %s. Data are corrupted." % ( phafile ))
    
    #Keep a copy which will never be modified
    self.counts_backup        = numpy.array(self.counts,copy=True)
    self.bkgCounts_backup     = numpy.array(self.bkgCounts,copy=True)
    
    #Effective area correction is disabled by default, i.e.,
    #the nuisance parameter is fixed to 1    
    self.nuisanceParameters       = {}
    self.nuisanceParameters['InterCalib'] = Parameter("InterCalib",1,0.9,1.1,0.01,fixed=True,nuisance=True)
    
  pass
  
  def useIntercalibrationConst(self,factorLowBound=0.9,factorHiBound=1.1):
    self.nuisanceParameters['InterCalib'].free()
    self.nuisanceParameters['InterCalib'].setBounds(factorLowBound,factorHiBound)
    
    #Check that the parameter is within the provided bounds
    value                     = self.nuisanceParameters['InterCalib'].value
    
    if( value < factorLowBound ):
            
      warnings.warn("The intercalibration constant was %s, lower than the provided lower bound." %(value,factorLowBound) +
                    " Setting it equal to the lower bound")
      
      self.nuisanceParameters['InterCalib'].setValue(float(factorLowBound))

      
      
    if( value > factorHiBound):
      
      warnings.warn("The intercalibration constant was %s, larger than the provided hi bound." %(value,factorHiBound) +
                    " Setting it equal to the hi bound")
      
      self.nuisanceParameters['InterCalib'].setValue(float(factorHiBound))
    
  def fixIntercalibrationConst(self,value=None):
    
    if(value is not None):
      #Fixing the constant to the provided value
      self.nuisanceParameters['InterCalib'].setValue(float(value))
    
    else:
    
      #Do nothing, i.e., leave the constant to the value
      #it currently has
      pass
      
    self.nuisanceParameters['InterCalib'].fix()
  
  def setActiveMeasurements(self,*args):
    '''Set the measurements to be used during the analysis. 
    Use as many ranges as you need,
    specified as 'emin-emax'. Energies are in keV. Example:
    
    setActiveMeasurements('10-12.5','56.0-100.0')
    
    which will set the energy range 10-12.5 keV and 56-100 keV to be 
    used in the analysis'''
    
    #To implelemnt this we will use an array of boolean index, 
    #which will filter
    #out the non-used channels during the logLike
    
    #Now build the mask: values for which the mask is 0 will be masked
    mask                      = numpy.zeros(self.phafile.getRates().shape)
    
    for arg in args:
      ee                      = map(float,arg.replace(" ","").split("-"))
      emin,emax               = sorted(ee)
      idx1                    = self.response.energyToChannel(emin)
      idx2                    = self.response.energyToChannel(emax)
      mask[idx1:idx2+1]       = True
    pass
    self.mask                 = numpy.array(mask,numpy.bool)
    
    self.counts               = self.counts_backup[self.mask]
    self.bkgCounts            = self.bkgCounts_backup[self.mask]
    
    print("Now using %s channels out of %s" %( numpy.sum(self.mask),
                                               self.phafile.getRates().shape[0]
                                              ) )
  pass
  
  def getName(self):
    '''
    Return a name for this dataset (likely set during the constructor)
    '''
    return self.name
  pass
  
  def setModel(self,likelihoodModel):
    '''
    Set the model to be used in the joint minimization.
    '''
    self.likelihoodModel         = likelihoodModel
    
    nPointSources                = self.likelihoodModel.getNumberOfPointSources()
    
    #This is a wrapper which iterates over all the point sources and get
    #the fluxes
    #We assume there are no extended sources, since the GBM cannot handle them
    
    def diffFlux(energies):
      
      fluxes                     = self.likelihoodModel.getPointSourceFluxes(0,energies)
      
      #If we have only one point source, this will never be executed
      for i in range(1, nPointSources):
        fluxes                  += self.likelihoodModel.getPointSourceFluxes(i,energies)
      
      return fluxes
    
    self.diffFlux                = diffFlux
    
    #The following integrates the diffFlux function using Simpson's rule
    #This assume that the intervals e1,e2 are all small, which is guaranteed
    #for any reasonable response matrix, given that e1 and e2 are Monte-Carlo
    #energies. It also assumes that the function is smooth in the interval
    #e1 - e2 and twice-differentiable, again reasonable on small intervals for
    #decent models. It might fail for models with too sharp features, smaller
    #than the size of the monte carlo interval.
    
    def integral(e1,e2):
      
      #Simpson's rule
      
      return (e2 - e1) / 6.0 * ( self.diffFlux(e1) 
                                 + 4 * self.diffFlux( (e1+e2) / 2.0 )
                                 + self.diffFlux(e2) )
    
    self.response.setFunction( diffFlux,
                               integral)
  pass

  def innerFit(self):
        
    #Effective area correction
    if(self.nuisanceParameters['InterCalib'].isFree()):
            
      #A true fit would be an overkill, and slow
      #Just sample a 100 values and choose the minimum
      values                  = numpy.linspace(self.nuisanceParameters['InterCalib'].minValue,
                                               self.nuisanceParameters['InterCalib'].maxValue,
                                               100)
      
      
      #I do not use getLogLike so I can compute only once the folded model
      #(which is not going to change during the inner fit)
      
      folded                  = self.getFoldedModel()
      
      modelCounts             = folded * self.exposure
      
      def fitfun(cons):
        
        self.nuisanceParameters['InterCalib'].setValue( cons )
        
        return (-1) * self._computeLogLike(self.nuisanceParameters['InterCalib'].value * modelCounts + self.bkgCounts)
            
      logLval                 = map(fitfun, values)
      idx                     = numpy.argmax(logLval)
      self.nuisanceParameters['InterCalib'].setValue(values[idx])
      #return logLval[idx]
      
      #Now refine with minuit
      
      parameters              = collections.OrderedDict()
      parameters[ (self.name, 'InterCalib') ]      = self.nuisanceParameters['InterCalib']
      minimizer               = minimization.iMinuitMinimizer(fitfun, parameters)
      bestFit, mlogLmin       = minimizer.minimize()
      
      return mlogLmin * (-1)
      
    else:
      
      return self.getLogLike()
  
  def getFoldedModel(self):
    
    #Get the folded model for this spectrum 
    #(this is the rate predicted, in cts/s)    
    
    return self.response.convolve()[self.mask]
  
  def getModelAndData(self):
    
    e1,e2                     = (self.response.ebounds[:,0],
                                 self.response.ebounds[:,1])
    
    return ( self.response.convolve()[self.mask] * self.exposure 
           + self.bkgCounts, 
             e1[self.mask],
             e2[self.mask],
             self.counts )
  
  def display( self ):
    
    #Plot the counts spectrum with residuals
    
    model, e1, e2, counts = self.getModelAndData()
    
    #Try to automagically decide a good constant number
    #of counts, based on the input
    
    total = numpy.sum( counts )
    
    trials = [ 50, 25, 15, 5]
    
    choice = None
    
    for t in trials:
        
        if total / float( t ) >= 10:
            
            choice = t
            
            break
    
    #If no choice worked, use the minimum, which is 5 counts
    #per bin
    
    if choice is None:
        
        choice = 5
    
    #Bin the data by constant counts
    
    binner = Binner.Binner( e1, e2, counts )
    
    ne1, ne2, nc, newModel = binner.byConstantCounts( choice, model )
        
    #Now plot the results
    
    fig = plt.figure()
    
    gs = gridspec.GridSpec(2,1, height_ratios=[2,1])
    gs.update(hspace=0)
    
    sub = plt.subplot( gs[0] )
    
    #Plot data
    
    ec = ( ne2 + ne1 ) / 2.0
    de = ( ne2 - ne1 ) / 2.0
    
    y = nc / ( de * 2.0 ) / self.exposure
    dy = numpy.sqrt( nc ) / ( de * 2.0 ) / self.exposure
    
    sub.errorbar( ec, y, xerr=de, yerr = dy, fmt='.', capsize=0 )
    
    #Plot model
    
    xx = numpy.append( ne1, [ ne2[-1] ])
    
    yy = newModel / ( 2.0 * de ) / self.exposure
    
    yyy = numpy.append( yy, yy[-1] )

    sub.step( xx, yyy, where='post' )
    
    sub.set_xscale("log")
    sub.set_yscale("log", nonposy='clip')
    
    sub.set_ylabel(r"Counts keV$^{-1}$ s$^{-1}$") 
    
    #Now plot residuals
    sub1 = plt.subplot( gs[1] )
    
    res = ( yy - y ) / dy
    
    sub1.axhline( 0, linestyle='--')
    sub1.errorbar( ec, res, yerr=1.0, xerr=de, fmt='.', capsize=0 )
    
    sub1.set_xscale("log")
    
    #Match the x axis
    
    sub.set_xticks([])
    
    sub.set_xlim( [( ec - de).min(), (ec + de).max() ] ) 
    sub1.set_xlim( [( ec - de).min(), (ec + de).max() ] ) 
    
    sub1.set_xlabel("Energy (keV)")
    
    return fig
  
  def _getModelCounts(self):
    
    #Get the folded model for this spectrum (this is the rate predicted, 
    #in cts/s)
    
    folded                    = self.getFoldedModel()
    
    #Model is folded+background (i.e., we assume negligible errors on the 
    #background)
    modelCounts               = self.nuisanceParameters['InterCalib'].value * folded * self.exposure + self.bkgCounts
    
    return modelCounts
  
  def _computeLogLike(self, modelCounts):
    
    idx = modelCounts > 0
    
    return numpy.sum(- modelCounts[idx] 
                     + self.counts[idx] * numpy.log( modelCounts[idx] )
                     - logfactorial( self.counts[idx] ) )
  
  def getLogLike(self):
    '''
    Return the value of the log-likelihood with the current values for the
    parameters
    '''
    
    modelCounts               = self._getModelCounts()
    
    logLike                   = self._computeLogLike( modelCounts )
    
    return logLike
      
  def getNuisanceParameters(self):
    '''
    Return a list of nuisance parameter names. Return an empty list if there
    are no nuisance parameters
    '''
    return self.nuisanceParameters.keys()
  pass
pass

class Response(object):
  def __init__( self, rspfile, arffile = None ):
    
    rspNumber                   = 1
    
    #Check if we are dealing with a .rsp2 file (containing more than
    #one response). This is checked by looking for the syntax 
    # [responseFile]{[responseNumber]}
    
    if('{' in rspfile):
    
      tokens                    = rspfile.split("{")
      rspfile                   = tokens[0]
      rspNumber                 = int(tokens[-1].split('}')[0].replace(" ",""))
    
    
    #Read the response
    with pyfits.open(rspfile) as f:
      
      try:
        
        #GBM typical response
        data                    = f['MATRIX',rspNumber].data
      
      except:
        #Other detectors might use the SPECRESP MATRIX name instead
        #Note that here we are not catching any exception, because
        #we have to fail if we cannot read the matrix
        
        data                    = f['SPECRESP MATRIX',rspNumber].data
      
      #Sometimes .rsp files contains a weird format featuring variable-length
      #arrays. Historically those confuse pyfits quite a lot, so we ensure 
      #to transform them into standard numpy matrices to avoid issues
      
      self.matrix               = variableToMatrix(data.field('MATRIX'))
      
      self.ebounds              = numpy.vstack([ f['EBOUNDS'].data.field("E_MIN"),
                                                 f['EBOUNDS'].data.field("E_MAX") ]).T
      
      self.mc_channels          = numpy.vstack([ data.field("ENERG_LO"),
                                                 data.field("ENERG_HI") ]).T
      
      #Now let's see if we have a ARF, if yes, read it
      
      if arffile is not None:
          
          with pyfits.open( arffile ) as f:
              
              data = f['SPECRESP'].data
              
          
          arf = data.field('SPECRESP')
          
          #Check that arf and rmf have same dimensions
          
          if arf.shape[0] != self.matrix.shape[1]:
              
              raise IOError("The ARF and the RMF file does not have the same number of channels")
          
          #Check that the ENERG_LO and ENERG_HI for the RMF and the ARF
          #are the same
          
          arf_mc_channels = numpy.vstack([ data.field("ENERG_LO"),
                                           data.field("ENERG_HI") ]).T
          
          #Declare the mc channels different if they differ by more than
          #1%
          
          idx = ( self.mc_channels > 0 )
          
          diff = ( self.mc_channels[idx] - arf_mc_channels[idx] ) / self.mc_channels[idx]
          
          if diff.max() > 0.01:
             
             raise IOError("The ARF and the RMF have one or more MC channels which differ by more than 1%")
          
          #Multiply ARF and RMF
          
          self.matrix = self.matrix * arf
          
    
  def setFunction(self, differentialFunction, integralFunction=None):
    '''
    Set the function to be used for the convolution
    '''
    
    self.differentialFunction   = differentialFunction
    
    if(integralFunction==None):
      
      #This should never happen in 3ML, but we keep this functionality for
      #other uses. Print a warning anyway
      
      warnings.warn("Using the slow numerical integration in the GBM plugin!", 
                     RuntimeWarning)
      
      def integral(x,y):
        return scipy.integrate.quad(self.differentialFunction,x,y)[0]
      
      #NB: vectorize is super slow!
      self.integralFunction     = numpy.vectorize(integral,otypes=[numpy.float])
    
    else:
      
      self.integralFunction     = integralFunction
    
  
    
  def convolve(self):
    
    trueFluxes                  = self.integralFunction( self.mc_channels[:,0],
                                                         self.mc_channels[:,1] )
    
    #Sometimes some channels have 0 lenths, or maybe they start at 0, where
    #many functions (like a power law) are not defined. In the response these
    #channels have usually a 0, but unfortunately for a computer
    # inf * zero != zero. Thus, let's force this. We avoid checking this situation
    #in details because this would have a HUGE hit on performances
    
    idx = numpy.isfinite( trueFluxes )
    trueFluxes[ ~idx ] = 0    
    
    foldedCounts                = numpy.dot( trueFluxes, self.matrix.T )
    
    return foldedCounts
  
  def getCountsVector(self,e1,e2):
    
    trueFluxes                  = self.integralFunction(self.mc_channels[:,0],self.mc_channels[:,1])
  
  def energyToChannel(self,energy):
    
    '''Finds the channel containing the provided energy. 
    NOTE: returns the channel index (starting at zero),
    not the channel number (likely starting from 1)'''
    
    #Get the index of the first ebounds upper bound larger than energy
    
    try:
    
      idx                         = next( idx for idx,
                                          value in enumerate(self.ebounds[:,1]) 
                                          if value >= energy )
    
    except StopIteration:
      
      #No values above the given energy, return the last channel
      return self.ebounds[:,1].shape[0]
    
    return idx
  

def variableToMatrix(variableLengthMatrix):
  '''This take a variable length array and return it in a 
  properly formed constant length array, to avoid some pyfits obscure bugs'''
  
  nrows                          = len(variableLengthMatrix)
  ncolumns                       = max([len(elem) for elem in variableLengthMatrix])
  matrix                         = numpy.zeros([ncolumns,nrows])
  
  for i in range(nrows):
    for j in range(ncolumns):
      
      try:
      
        matrix[j,i]                = variableLengthMatrix[i][j]
      
      except:
        
        #This happens when the i,j element does not exist, which is not an error
        #We will just leave it to zero
        pass
        
  return matrix



# Author: G.Vianello (giacomov@stanford.edu)

import collections


class DataList(object):
    """
    A container for data sets. Can be accessed as a dictionary,
    with the [key] operator.
    """

    def __init__(self, *data_sets):

        self._innerDict = collections.OrderedDict()

        for d in data_sets:

            if d.getName() in self._innerDict.keys():

                raise RuntimeError("You have to use unique names for data sets. %s already exists." % (d.getName()))

            else:

                self._innerDict[d.getName()] = d

    def __setitem__(self, key, value):

        # Enforce the unique name
        if key in self.keys():

            raise RuntimeError("You have to use unique names for data sets. %s already exists." % key)

        else:

            self._innerDict[key] = value

    def __getitem__(self, key):

        return self._innerDict[key]

    def keys(self):

        return self._innerDict.keys()

    def values(self):

        return self._innerDict.values()


pass



from threeML.models.spectralmodel import SpectralModel
from threeML.models.Parameter import Parameter
import math
import scipy.integrate
import operator

import collections



class Blackbody(SpectralModel):

    def setup(self):
        self.functionName        = "Blackbody"
        self.formula             = r'$f(E) =  A E^{2}\frac{1}{{\rm exp} (E/kT) -1}$'
        self.parameters          = collections.OrderedDict()
        self.parameters['A'] = Parameter('A',100.,1.,1E6,0.1,fixed=False,nuisance=False,dataset=None,normalization=True)
        self.parameters['kT'] = Parameter('kT',30.,1.,1E6,0.1,fixed=False,nuisance=False,dataset=None)
        
        
            
        Self.ncalls              = 0
    
#        def integral(e1,e2):
#            # No analytic expression
#            pass
#            
#        self.integral            = None
 
  
    def __call__(self,energy):
        self.ncalls             += 1
        A                        = self.parameters['A'].value
        kT                       = self.parameters['kT'].value
        
        return numpy.maximum( A*energyx**2.*power( exp(energy/kT) -1., -1.), 1e-30)
   
  
    def photonFlux(self,e1,e2):
        return self.integral(e1,e2)
  
    def energyFlux(self,e1,e2):
        pass


        #return (eF(e2)-eF(e1))*keVtoErg




import astropy.io.fits as fits
import scipy.interpolate
import numpy
from threeML.models.spectralmodel import SpectralModel
import copy
class TableModel(SpectralModel):

    
    def SetTableFile(self,tableFile):

    
        self._ReadTable(tableFile)
        self._CreateInterpolation()
        

    
        
    def _ReadTable(self,tableFile):
        '''
        Virtual function. Implementation defined
        later.
        
        1) Should create an array called self._tableParams
        2) Create an array of evaluation energies
        3) Create a matrix of fluxes of shape (n_1,n_2,...n_i,j)
            where n are the number of values for the ith parameter
            and j is the number of evaluation energies
        
        '''
                
        pass
    
    def _CreateInterpolation(self):
        
        # Create the interpolation grid
        # 
        tmp = copy.deepcopy(self._tableParams)
        
        tmp.append(self._evalEnergies.tolist())
        tmp = map(lambda x: numpy.array(x,dtype=numpy.float32),tmp)
        interpGrid = tuple(tmp)
        self._tableFluxes.dtype = numpy.float32
        zero = numpy.float32(0.)
        
        self._interpFunc = scipy.interpolate.RegularGridInterpolator(interpGrid,self._tableFluxes,method="linear",fill_value=zero,bounds_error=False)
        

        
        
class FITSTableModel(TableModel):
    
    def _ReadTable(self,fitsFileName):
        
        # Open the FITS file
        self._fitsFile = fits.open(fitsFileName)
        
        # Extract the evaluation energies
        self._ExtractEvalEnergies()
        
        # Extract the parameters
        self._ExtractParameters()
        
        # Extract the fluxes and reshape them
        # For the interpolator
        shape = self._numParamValues
        shape.append(self._numEvalEnergies)
        
        self._tableFluxes =  self._fitsFile[3].data['INTPSPEC'].reshape(*shape) / self._binWidths #convert!
        
        
        
        
    def _ExtractEvalEnergies(self):
    
        self._evalEnergies =  numpy.array(map(numpy.mean,self._fitsFile[2].data))
        self._binWidths = self._fitsFile[2].data["ENERG_HI"] - self._fitsFile[2].data["ENERG_LO"]
        
        self._numEvalEnergies = len(self._evalEnergies)
        
    def _ExtractParameters(self):
        
        self._numParamValues = (self._fitsFile[1].data["NUMBVALS"]).tolist()
        self._tableParams    = (self._fitsFile[1].data["VALUE"]).tolist()
    
class NumpyTableModel(TableModel):

    def _ReadTable(self,npzfile):

        npz = numpy.load(npzfile)
        self._evalEnergies = npz["energy"]
        self._tableParams  = npz["params"].tolist()
        self._tableFluxes  = npz["fluxes"]


def MakeNumpyTableModel(params,evaluationEnergies,tableFluxes,filename):
    '''
    Helper function to save a table model into a numpy format.

    params:              a list of lists for the parameters values
    evalulationEnergies: array of the energies the model is evaluated at
    tablefluxes:         the table fluxes in a nested array
    filename: File name to save the table to. No extension required

    An example of how to build the table and params::

    f(ene,par1,par2) ---> the generating function
    
    param1 = [1,2,3,...] ---> the parameter values for par1
    param2 = [5,6,7,...] ---> the parameter values for par2

    eneGrid = [1,....,1e5] ---> the evaluation energies

    tabMod = []
    for p1 in par1:
     tmp = []
     for p2 in par2:
      tmp2.append(f(eneGrid,p1,p2))
     tmp.append(tmp2)
    tabMod.append(tmp)

    params = [param1,param2]


    MakeNumpyTable(params,eneGrid,tabMod,"table")    
    
    '''

    #Check the shape
    tableShape = tableFluxes.shape
    tmp = []
    for p in params:
        
        tmp.append(len(p))
    tmp.append(len(evaluationEnergies))
    
    if tuple(tmp) != tableShape:
        print "The parameters, evaluation energies and"
        print "table fluxes do not have the same shape!"
        print tuple(tmp)
        print tableShape
    
        return
    f = open(filename,"w")
    
    numpy.savez_compressed(f,params=params,energy=evaluationEnergies,fluxes=tableFluxes )
    
    f.close()



from threeML.models.spectralmodel import SpectralModel
from threeML.models.Parameter import Parameter
import math
import scipy.integrate
import operator
#import numexpr
import numpy


import collections



class PowerLaw(SpectralModel):

    def setup(self):
        self.functionName        = "Powerlaw"
        self.formula             = r'\begin{equation}f(E) = A (E / E_{piv})^{\gamma}\end{equation}'
        self.parameters          = collections.OrderedDict()
        self.parameters['gamma'] = Parameter('gamma',-2.0,-10,10,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['logA']     = Parameter('logA',-4,-40,30,0.1,fixed=False,nuisance=False,dataset=None,normalization=False)
        self.parameters['Epiv']  = Parameter('Epiv',1.0,1e-10,1e10,1,fixed=True,unit='keV')
    
        self.ncalls              = 0
    
        def integral(e1,e2):
            a                      = self.parameters['gamma'].value
            piv                    = self.parameters['Epiv'].value
            norm                   = pow(10, self.parameters['logA'].value)
      
            if(a!=-1):
                def f(energy):
                    return norm * energy * math.pow(energy/piv,a)/(a+1)
            else:
                def f(energy):
                    return norm * piv * math.log(energy)
                    
            return f(e2)-f(e1)
        self.integral            = integral
 
  
    def __call__(self,e):
        self.ncalls             += 1
        piv                      = self.parameters['Epiv'].value
        norm                     = pow(10, self.parameters['logA'].value)
        gamma                    = self.parameters['gamma'].value
        
        energies = numpy.array( e, ndmin=1, copy=False)

        return numpy.maximum( norm * (energies/piv)**gamma, 1e-100)
   
  
    def photonFlux(self,e1,e2):
        return self.integral(e1,e2)
  
    def energyFlux(self,e1,e2):
        a                        = self.parameters['gamma'].value
        piv                      = self.parameters['Epiv'].value
        norm                     = pow(10, self.parameters['logA'].value)
        if(a!=-2):
            def eF(e):
                return numpy.maximum( norm * numpy.power(e/piv,2-a)/(2-a), 1e-100 )
        else:
            def eF(e):
                return numpy.maximum( norm * numpy.log(e/piv), 1e-100 )
   
    
        return (eF(e2)-eF(e1))*self.keVtoErg




from powerlaw import PowerLaw
from scipy import weave
from scipy.weave import converters

class wPowerLaw(PowerLaw):
    def __call__(self,energy):
        a                         = self.parameters['A'].value
        alpha                     = self.parameters['gamma'].value
        try:
            n                       = energy.shape[0]
        except:
            n                       = 1
    
        result                    = numpy.array([0]*n,dtype=float)
        code="""
        int i;
        for(i=0;i<n;i++)
        {
        result(i) = a*pow(energy(i),alpha);
        }
        """
        weave.inline(code,['a','alpha','n','energy','result'],type_converters=converters.blitz)
        return result





import numpy
import abc
import math

import scipy.stats

class Prior(object):
  __metaclass__               = abc.ABCMeta
    
  @abc.abstractmethod
  def setBounds(self,newMinValue,newMaxValue):
    '''
    Set the minimum and maximum values for the parameter
    '''
    pass
  pass
  
  @abc.abstractmethod
  def getName(self):
    '''
    Return the name of the prior
    '''
    pass
  pass 
  
  @abc.abstractmethod
  def __call__(self,value):
    '''
    Return the logarithm of the prior pdf at the given value
    '''
    pass
  pass
  
class UniformPrior(Prior):
  def __init__(self,minValue,maxValue):
    self.minValue             = minValue
    self.maxValue             = maxValue
  pass
  
  def getName(self):
    return "UniformPrior"
  
  def setBounds(self,newMinValue,newMaxValue):
    self.minValue             = newMinValue
    self.maxValue             = newMaxValue
  pass
  
  def __call__(self,value):
    if(self.minValue < value < self.maxValue):
      return 0.0
    else:
      return -numpy.inf
    pass
  pass
  
  def multinestCall(self,cube):
    return cube * (self.maxValue - self.minValue ) + self.minValue
pass

class LogUniformPrior(Prior):
  def __init__(self,minValue,maxValue):
    self.minValue             = minValue
    self.maxValue             = maxValue
  pass
  
  def getName(self):
    return "LogUniformPrior"
  
  def setBounds(self,newMinValue,newMaxValue):
    self.minValue             = newMinValue
    self.maxValue             = newMaxValue
  pass
  
  def __call__(self,value):
    if(self.minValue < value < self.maxValue and value>0):
      #This is = log(1/value)
      return -math.log10(value)
    else:
      return -numpy.inf
    pass
  pass
  
  def multinestCall(self,cube):
    decades                   = math.log1010(self.maxValue)-math.log1010(self.minValue)
    startDecade               = math.log1010(self.minValue)
    return 10**((cube * decades) + startDecade)

pass

class GaussianPrior(Prior):
  
  def __init__(self, mu, sigma):
    
    self.mu = float(mu)
    self.sigma = abs( float(sigma) )
        
    
  def getName(self):
    
    return "GaussianPrior"
  
  def setBounds(self,newMinValue,newMaxValue):
    #Useless in this context
    pass
  
  def __call__(self, value):
    
    #Here I multiply by sigma because of scipy "scale" behavior
    #(see scipy documentation)
    
    val                       = scipy.stats.norm.pdf( value, loc=self.mu, scale=self.sigma ) * self.sigma
  
    if( val < 1e-15):
      
      return - numpy.inf
    
    else:
      
      return math.log10(val)



import dill
import yaml
import os

from threeML.config.Config import threeML_config

try:
    
    from IPython.parallel import Client

except:
    
    from ipyparallel import Client
    
    
from contextlib import contextmanager

@contextmanager
def parallel_computation( profile=None ):
    
    #Memorize the state of the use-parallel config
    old_state = bool( threeML_config['parallel']['use-parallel'] )
    
    old_profile = str( threeML_config['parallel']['IPython profile name'] )
    
    #Set the use-parallel feature on
    threeML_config['parallel']['use-parallel'] = True
    
    #Now use the specified profile (if any), otherwise the default one
    if profile is not None:
        
        threeML_config['parallel']['IPython profile name'] = str( profile )
    
    #Here is where the content of the with parallel_computation statement gets
    #executed
    try:
    
        yield
    
    finally:
        
        #This gets executed in any case, even if there is an exception
        
        #Revert back
        threeML_config['parallel']['use-parallel'] = old_state
        
        threeML_config['parallel']['IPython profile name'] = old_profile

class ParallelClient(Client):
  
  def __init__(self, *args, **kwargs):
    
    #Just a wrapper around the IPython Client class
    #forcing the use of dill for object serialization
    #(more robust, and allows for serialization of class
    #methods)
    
    if 'profile' not in kwargs.keys():
        
        kwargs['profile'] = threeML_config['parallel']['IPython profile name']
    
    super(ParallelClient, self).__init__(*args, **kwargs)
    
    #This will propagate the use_dill to all running
    #engines    
    self.direct_view().use_dill()
 
  def getNumberOfEngines( self ):
    
    return len( self.direct_view() )



import numpy

from VirtualObservatoryCatalog import VirtualObservatoryCatalog

class GBMBurstCatalog(VirtualObservatoryCatalog):
    
    def __init__(self):
        
        super(GBMBurstCatalog, self).__init__('fermigbrst', 
                           'http://heasarc.gsfc.nasa.gov/cgi-bin/vo/cone/coneGet.pl?table=fermigbrst&',
                           'Fermi/GBM burst catalog')

    def applyFormat(self, votable):
        
        table = votable.to_table() 
        table['ra'].format = '5.3f'
        table['dec'].format = '5.3f'
        return table

#########

threefgl_types = {
'agn' : 'other non-blazar active galaxy',
'bcu' : 'active galaxy of uncertain type',
'bin' : 'binary',
'bll' : 'BL Lac type of blazar',
'css' : 'compact steep spectrum quasar',
'fsrq' : 'FSRQ type of blazar',
'gal' : 'normal galaxy (or part)',
'glc' : 'globular cluster',
'hmb' : 'high-mass binary',
'nlsy1' : 'narrow line Seyfert 1',
'nov' : 'nova',
'PSR' : 'pulsar, identified by pulsations',
'psr' : 'pulsar, no pulsations seen in LAT yet',
'pwn' : 'pulsar wind nebula',
'rdg' : 'radio galaxy',
'sbg' : 'starburst galaxy',
'sey' : 'Seyfert galaxy',
'sfr' : 'star-forming region',
'snr' : 'supernova remnant',
'spp' : 'special case - potential association with SNR or PWN',
'ssrq' : 'soft spectrum radio quasar',
'' : 'unknown'
}

class LATSourceCatalog(VirtualObservatoryCatalog):
    
    def __init__(self):
        
        super(LATSourceCatalog, self).__init__('fermilpsc', 
                           'http://heasarc.gsfc.nasa.gov/cgi-bin/vo/cone/coneGet.pl?table=fermilpsc&',
                           'Fermi/LAT source catalog')        

    def applyFormat(self, votable):
        
        table = votable.to_table() 
        table['ra'].format = '5.3f'
        table['dec'].format = '5.3f'
        table['Search_Offset'].format = '5.3f'
        
        def translate(key):
            if(key.lower()=='psr'):
                return threefgl_types[key]
            else:
                return threefgl_types[key.lower()]
        
        #Translate the 3 letter code to a more informative category, according
        #to the dictionary above       
        
        table['source_type'] = numpy.array(map(translate, table['source_type']))
                                
        new_table = table['name',
                          'source_type',
                          'ra','dec',
                          'assoc_name_1',
                          'tevcat_assoc',
                          'Search_Offset']
                
        return new_table.group_by('Search_Offset')
 



from threeML.minimizer import minimization
from threeML.exceptions import CustomExceptions
from threeML.io.Table import Table
from threeML.utils.cartesian import cartesian
from threeML.parallel.ParallelClient import ParallelClient
from threeML.io.ProgressBar import ProgressBar
from threeML.config.Config import threeML_config


import collections
import warnings
import copy

import numpy
import scipy.optimize
import scipy.stats
import sys, time
import matplotlib.pyplot as plt

import matplotlib.cm as cm
from matplotlib.colors import BoundaryNorm

from IPython.display import display

import numpy as np

class JointLikelihood(object):
  
  def __init__(self, likelihoodModel, dataList, **kwargs):
    
    #Process optional keyword parameters
    self.verbose             = False
        
    for k,v in kwargs.iteritems():
      
      if(k.lower()=="verbose"):
      
        self.verbose           = bool(kwargs["verbose"])
    
        
    self.likelihoodModel      = likelihoodModel
        
    self.dataList             = dataList
    
    for dataset in self.dataList.values():
      
      dataset.setModel(self.likelihoodModel)
    
    #This is to keep track of the number of calls to the likelihood
    #function
    self.ncalls              = 0
    
    self.setMinimizer('iMinuit')
    
  pass
  
  def preFit(self):
    
    #This is a simple scan through the normalization parameters to guess
    #a good starting point for them (if we start too far from the data,
    #minuit and the other minimizers will have trouble)
    
    #Get the list of free parameters
    freeParams               = self.likelihoodModel.getFreeParameters()
    
    #Now isolate the normalizations, and use them as free parameters in the loglikelihood
    
    self.freeParameters      = collections.OrderedDict()
    
    for (k,v) in freeParams.iteritems():
      
      if v.isNormalization():
        
        self.freeParameters[k] = v
        
    #Prepare the grid of values to scan
    
    grids                    = []
    
    for norm in self.freeParameters.values():
            
      grids.append(numpy.logspace( numpy.log10( norm.minValue ),
                                   numpy.log10( norm.maxValue ), 
                                   50 ))
    
    if len(grids) == 0:
        
        #No norm. Maybe they are fixed ?
        
        return
    
    #Compute the global likelihood at each point in the grid
    globalGrid               = cartesian(grids)
    
    logLikes                 = map(self.minusLogLikeProfile, globalGrid)
    
    idx                      = numpy.argmin(logLikes)
    #print("Minimum is %s with %s" %(logLikes[idx],globalGrid[idx]))
    
    for i,norm in enumerate(self.freeParameters.values()):
      
      norm.setValue(globalGrid[idx][i])
      norm.setDelta(norm.value / 40)
  
  def minusLogLikeProfile(self, *trialValues):
      
      #Keep track of the number of calls
      self.ncalls                += 1
      
      #Check that we don't have nans in the trialValues
      #(sometimes happen when MINUIT fails to estimate the derivatives)
      
      trialValues = numpy.array( trialValues )
      
      #This is the fastest way to check for any nan
      #(try other methods if you don't believe me)
      if not numpy.isfinite(numpy.dot(trialValues,trialValues.T)):
          
          return minimization.FIT_FAILED
      
      #Assign the new values to the parameters
      
      for i,parname in enumerate(self.freeParameters.keys()):
        
        self.likelihoodModel.parameters[parname[0]][parname[1]].setValue(trialValues[i])
      
            
      #Now profile out nuisance parameters and compute the new value 
      #for the likelihood
      
      globalLogLike              = 0
      
      for dataset in self.dataList.values():
          
          try:
            
            thisLogLike = dataset.innerFit()
          
          except CustomExceptions.ModelAssertionViolation:
            
            #This is a zone of the parameter space which is not allowed. Return
            #a big number for the likelihood so that the fit engine will avoid it
            warnings.warn("Fitting engine in forbidden space: %s" %(trialValues,))
            return minimization.FIT_FAILED
          
          except:
            
            #Do not intercept other errors
            raise
          
                      
          #if thisLogLike > 0:
             
          #    raise RuntimeError("LogLike cannot be larger than 0")
          
          globalLogLike         += thisLogLike
          
          #dataset.getLogLike()      
      
      if("%s" % globalLogLike=='nan'):
        warnings.warn("These parameters returned a logLike = Nan: %s" %(trialValues,))
        return minimization.FIT_FAILED
      
      if(self.verbose):
        print("Trying with parameters %s, resulting in logL = %s" %(trialValues,globalLogLike))
      
      return globalLogLike*(-1)
  
  
  def _setupMinimizer(self,minimizer):
  
    if(minimizer.upper()=="IMINUIT"):
    
      return minimization.iMinuitMinimizer
    
    elif(minimizer.upper()=="MINUIT"):
    
      return minimization.MinuitMinimizer
    
    elif(minimizer.upper()=="SCIPY"):
    
      return minimization.ScipyMinimizer
    
    elif(minimizer.upper()=="BOBYQA"):
    
      return minimization.BOBYQAMinimizer
    
    else:
    
      raise ValueError("Do not know minimizer %s" %(minimizer))
  
  def setMinimizer(self, minimizer):
    
    self.Minimizer = self._setupMinimizer(minimizer)
  
  def _getTableOfParameters(self,parameters):
    
    data = []
    nameLength = 0
    
    for k,v in parameters.iteritems():
     
     curName                 = "%s_of_%s" %(k[1],k[0])
               
     data.append([curName,"%s" % v.value,v.unit])
     
     if(len(v.name) > nameLength):
       nameLength = len(curName)
   
    pass
    
    table                     = Table(rows = data,
                                      names = ["Name","Value","Unit"],
                                      dtype=('S%i' %nameLength, str, 'S15'))
    
    return table

  
  def fit(self,prefit=False):
    
    #Pre-fit: will fit the normalizations so that they are not too far
    #from the data (which would make the fitting below fail)
    
    if prefit:
        
        self.preFit()
    
    #Isolate the free parameters
    #NB: nuisance parameters are NOT in this dictionary
    
    self.freeParameters       = self.likelihoodModel.getFreeParameters()
    
    #Now check and fix if needed all the deltas of the parameters
    #to 10% of their value (otherwise the fit will be super-slow)
    for k,v in self.freeParameters.iteritems():
      
      if (abs(v.delta) < abs(v.value) * 0.1):
                
        v.setDelta(abs(v.value) * 0.1)
    
    #Instance the minimizer
    self.minimizer            = self.Minimizer(self.minusLogLikeProfile,
                                               self.freeParameters)
    
    #Perform the fit
    xs, logLmin          = self.minimizer.minimize()
    
    #Print results
    print("Best fit values:\n")
    
    self.minimizer.printFitResults()
    
    print("Nuisance parameters:\n")
    
    nuisanceParam = collections.OrderedDict()
    
    for dataset in self.dataList.values():
      
      for pName in dataset.getNuisanceParameters():
        
        nuisanceParam[ ( dataset.getName(), pName )] = dataset.nuisanceParameters[pName]
    
    table = self._getTableOfParameters(nuisanceParam)
    display(table)
    
    print("\nCorrelation matrix:\n")
    
    self.minimizer.printCorrelationMatrix()
    
    print("\nMinimum of -logLikelihood is: %s\n" %(logLmin))
    
    print("Contributions to the -logLikelihood at the minimum:\n")
    
    mLogLikes                 = collections.OrderedDict()
    mLogLikes['total']        = logLmin
    
    self.currentMinimum       = float(logLmin)
    
    data                      = []
    nameLength                = 0
    
    for dataset in self.dataList.values():
      
      ml                      = dataset.getLogLike()*(-1)
      
      mLogLikes[dataset.getName()] = ml
      
      nameLength              = max(nameLength, len(dataset.getName()) + 1)
      data.append( [ dataset.getName(), ml ] )
    
    table                     = Table( rows  = data,
                                       names = ["Detector","-LogL"],
                                       dtype = ('S%i' %nameLength, float))
          
    table['-LogL'].format = '2.2f'
    
    display(table)
    
    #Prepare the dictionary with the results
    results                   = collections.OrderedDict()
    
    results['parameters']     = xs
    results['minusLogLike']   = mLogLikes
    
    return results
  
  def getErrors(self,fast=False):
    
    if(not hasattr(self,'currentMinimum')):
      raise RuntimeError("You have to run the .fit method before calling errors.")
        
    return self.minimizer.getErrors(fast)
    
  def getLikelihoodProfiles(self):
    
    if(not hasattr(self,'minimizer')):
      raise RuntimeError("You have to run the .fit method before calling errors.")
    
    return self.minimizer.getLikelihoodProfiles()
  
  def getContours(self, src1, param1, p1min, p1max, p1steps,
                        src2=None, param2=None, p2min=None, p2max=None, p2steps=None,
                        **kwargs):
    
    if(not hasattr(self,'currentMinimum')):
      raise RuntimeError("You have to run the .fit method before calling getContours.")
    
    
    if param1==param2 and src1==src2:
      
      raise RuntimeError("You have to specify two different parameters.")
    
    if p1min > p1max:
      
      raise RuntimeError("Minimum larger than maximum for parameter 1")
    
    if (not src2 is None) and p2min > p2max:
      
      raise RuntimeError("Minimum larger than maximum for parameter 2")
    
    #Default values
    debug = False
    
    #Check the current keywords
    for k,v in kwargs.iteritems():
        
        if k=="debug":
            
            debug = bool( v )
    
    
    #Check that parameters are existent and free
    for s,p in zip( [ src1, src2 ], [ param1, param2 ] ):
      
      if s is None:
          
          #Stepping through one parameter only
          continue
      
      if ( (s, p) not in self.freeParameters.keys() ):
           
           raise ValueError("Parameter %s of source %s is not a free parameter of current model" %(p,s))
    
    if not threeML_config['parallel']['use-parallel']:
      
      #Create a new minimizer to avoid messing up with the best
      #fit
      
      #Copy of the parameters
      backup_freeParameters = copy.deepcopy(self.freeParameters)
      
      minimizer = self.Minimizer(self.minusLogLikeProfile,
                                 self.freeParameters)
      
      a, b, cc = minimizer.contours(src1, param1, p1min, p1max, p1steps,
                                    src2, param2, p2min, p2max, p2steps,
                                    True, **kwargs)
      
      #Restore the original
      self.freeParameters = backup_freeParameters
      
      if src2 is None:
      
          cc = cc[:, 0]
                                             
    else:
      
      #With parallel computation
      
      #Connect to the engines
      try:
        
        client = ParallelClient(**kwargs)
      
      except:
        
        sys.stderr.write("\nCannot connect to IPython cluster. Is the cluster running ?\n\n\n")
        
        raise RuntimeError("Cannot connect to IPython cluster.")

      
      threads = client.getNumberOfEngines()
      
      if( threads > p1steps):
        
        threads = int(p1steps)
        
        warnings.warn("The number of threads is larger than the number of steps. Reducing it to %s." %(threads))
                  
      #Check if the number of steps is divisible by the number
      #of threads, otherwise issue a warning and make it so
      
      if( float(p1steps) % threads != 0 ):
        
        p1steps = p1steps // threads * threads
        
        warnings.warn("The number of steps is not a multiple of the number of threads. Reducing steps to %s" %(p1steps))
      
      #Now this is guaranteed to be a integer number
      p1_split_steps = p1steps // int(threads)
      
      #Prepare arrays for results
      
      if src2 is None:
          
          #One array
          pcc = pcc = numpy.zeros( ( p1steps ) )
          
          pa = numpy.linspace( p1min, p1max, p1steps )
          pb = None
      
      else:
      
          pcc = numpy.zeros( ( p1steps, p2steps ) )
      
          #Prepare the two axes of the parameter space
          pa = numpy.linspace( p1min, p1max, p1steps )
          pb = numpy.linspace( p2min, p2max, p2steps )
      
      #Define the parallel worker which will go through the computation
      
      #NOTE: I only divide
      #on the first parameter axis so that the different
      #threads are more or less well mixed for points close and
      #far from the best fit
      
      def worker(i):
        
        #Re-create the minimizer
        
        #backup_freeParameters = copy.deepcopy(self.freeParameters)
        
        minimizer = self.Minimizer(self.minusLogLikeProfile,
                                   self.freeParameters)
        
        this_p1min = pa[i * p1_split_steps]
        this_p1max = pa[(i+1) * p1_split_steps - 1]
        
        if debug:
            
            print("From %s to %s" %(this_p1min, this_p1max))
        
        aa, bb, ccc = minimizer.contours(src1, param1, this_p1min, this_p1max, p1_split_steps,
                                         src2, param2, p2min, p2max, p2steps,
                                         False, **kwargs)
        
        #self.freeParameters = backup_freeParameters
        
        return ccc
              
      lview = client.load_balanced_view()
      #lview.block = True
      amr = lview.map_async(worker, range(threads))
      
      #Execute and print progress
      
      prog = ProgressBar(threads)
      
      while not amr.ready():
        
        #Avoid checking too often
        time.sleep(1 + np.random.uniform(0,1))
        
        if(debug):
          stdouts = amr.stdout
          
          # clear_output doesn't do much in terminal environments
          for stdout, stderr in zip(amr.stdout, amr.stderr):
            if stdout:
                print "%s" % (stdout[-1000:])
            if stderr:
                print "%s" % (stderr[-1000:])
          sys.stdout.flush()        
        
        prog.animate( amr.progress - 1 )
      
      #Force to display 100% at the end
      prog.animate( threads - 1 )
      
      #Now get results and print some diagnostic
      print("\n")
      
      #print("Serial time: %1.f (speed-up: %.1f)" %(amr.serial_time, float(amr.serial_time) / amr.wall_time))
      
      res = amr.get()
      
      for i in range(threads):
        
        if src2 is None:
        
            pcc[ i * p1_split_steps : (i+1) * p1_split_steps ] = res[i][:,0]
        
        else:
        
            pcc[ i * p1_split_steps : (i+1) * p1_split_steps, : ] = res[i]
      
      #Keep them separated up to now for debugging purposes
      cc = pcc
      a = pa
      b = pb
      
    pass
    
    if src2 is not None:
    
        fig = self._plotContours("%s of %s" %(param1, src1), a, "%s of %s" % (param2, src2), b, cc)
    
    else:
        
        fig = self._plotProfile( "%s of %s" %(param1, src1), a, cc )
    
    #Check if we found a better minimum
    if( self.currentMinimum - cc.min() > 0.1 ):
      
      if src2 is not None:
      
          idx = cc.argmin()
          
          aidx, bidx = numpy.unravel_index(idx, cc.shape)
          
          print("\nFound a better minimum: %s with %s = %s and %s = %s" 
                 %(cc.min(),param1,a[aidx],param2,b[bidx]))
      
      else:
          
          idx = cc.argmin()
          
          print("Found a better minimum: %s with %s = %s" 
                 %(cc.min(),param1,a[idx]))
    
    pass
    
    return a, b, cc, fig
  
  def _plotProfile(self, name1, a, cc):
      
      #plot 1,2 and 3 sigma horizontal lines
      sigmas = [1,2,3]
      
      #Compute the corresponding probability. We do not
      #pre-compute them because we will introduce at
      #some point the possibility to personalize the
      #levels
      probs = []
      
      for s in sigmas:
          
          #One-sided probability
          #It is one-sided because we consider one side at the time
          #when computing the error
          
          probs.append( 1 - (scipy.stats.norm.sf(s) * 2) )
      
      #Compute the corresponding delta chisq. (chisq has 1 d.o.f.)
      
      deltachi2 = scipy.stats.chi2.ppf(probs, 1) / 2.0 #two-sided!
      
      fig = plt.figure()
      sub = fig.add_subplot(111)
      
      #Neutralize values of the loglike too high
      #(fit failed)
      idx = (cc == minimization.FIT_FAILED)
      
      sub.plot( a[~idx], cc[~idx], lw=2)
      
      #Now plot the failed fits as "x"
      
      sub.plot( a[idx], [cc.min()] * a[idx].shape[0], 'x', c='red',markersize=2)
      
      #Decide colors
      colors = ['blue','cyan','red']
      
      for s,d,c in zip( sigmas, deltachi2, colors ):
          
          sub.axhline( self.currentMinimum + d , linestyle='--' , 
                       color=c, label=r'%s $\sigma$' %( s ), lw = 2 )
      
      #Fix the axis to cover from the minimum to the 3 sigma line
      sub.set_ylim( [ self.currentMinimum - deltachi2[0] , 
                      self.currentMinimum + (deltachi2[-1] * 2) ] )
      
      
      plt.legend( loc=0, frameon=True )
      
      sub.set_xlabel( name1 )
      sub.set_ylabel( "-log( likelihood )" )
      
      return fig
      
  
  def _plotContours(self, name1, a, name2, b, cc):
      
      #Check that we have something to plot
      
      delta = cc.max() - cc.min()
      
      if delta < 0.5:
          
          print("\n\nThe maximum difference in statistic is %s among all the points in the grid." % delta)
          print(" This is too small. Enlarge the search region to display a contour plot")
          
          return None
      
      #plot 1,2 and 3 sigma contours
      sigmas = [1,2,3]
      
      #Compute the corresponding probability. We do not
      #pre-compute them because we will introduce at
      #some point the possibility to personalize the
      #levels
      probs = []
      
      for s in sigmas:
          
          #One-sided probability
          #It is one-sided because we consider one side at the time
          #when computing the error
          
          probs.append( 1 - (scipy.stats.norm.sf(s) * 2) )
      
      #Compute the corresponding delta chisq. (chisq has 2 d.o.f.)
      deltachi2 = scipy.stats.chi2.ppf(probs, 2) / 2.0 #two-sided
      
      #Boundaries for the colormap
      bounds = [self.currentMinimum]
      bounds.extend(self.currentMinimum + deltachi2)
      bounds.append(cc.max())
      
      #Define the color palette
      palette = cm.Pastel1
      palette.set_over('white')
      palette.set_under('white')
      palette.set_bad('white')
      
      fig = plt.figure()
      sub = fig.add_subplot(111)
      
      #Show the contours with square axis
      im = sub.imshow(cc,
                 cmap=palette,
                 extent=[b.min(),b.max(),a.min(),a.max()],
                 aspect=(b.max()-b.min())/(a.max()-a.min()),
                 origin='lower',
                 norm=BoundaryNorm(bounds,256),
                 interpolation='bicubic',
                 vmax=(self.currentMinimum+deltachi2).max())
      
      #Plot the color bar with the sigmas
      cb = fig.colorbar(im, boundaries=bounds[:-1])
      lbounds = [0]
      lbounds.extend(bounds[:-1])
      cb.set_ticks(lbounds)
      ll = ['']
      ll.extend(map(lambda x:r'%i $\sigma$' %x, sigmas))
      cb.set_ticklabels(ll)
      
      #Align the labels to the end of the color level
      for t in cb.ax.get_yticklabels():
          t.set_verticalalignment('baseline')   
      
      #Draw the line contours
      sub.contour(b,a, cc, self.currentMinimum + deltachi2)
      
      sub.set_xlabel(name2)
      sub.set_ylabel(name1)
      
      return fig
   
#  def _restoreBestFitValues(self):
#    #Restore best fit values
#    for k in self.freeParameters.keys():
#      self.freeParameters[k].setValue(self.bestFitValues[k])
#      self.modelManager[k].setValue(self.bestFitValues[k])
#    pass  
#  pass
  
#  def getErrors(self,confidenceLevel=0.68268949213708585,**kwargs):
#    '''
#    Compute asymptotic errors using the Likelihood Ratio Test. Usage:
#    
#    computeErrors(0.68)
#    
#    will compute the 1-sigma error region, while:
#    
#    computeErrors(0.99)
#    
#    will compute the 99% c.l. error region, and so on. Alternatively, you
#    can specify the number of sigmas corresponding to the desired c.l., as:
#    
#    computeErrors(sigma=1)
#    
#    to get the 68% c.l., or:
#    
#    computeErrors(sigma=2)
#    
#    to get the ~95% c.l.
#    '''
#    
#    equivalentSigma           = None
#    plotProfiles              = False
#    for k,v in kwargs.iteritems():
#      if(k.lower()=="sigma"):
#        equivalentSigma       = float(v)
#      elif(k.lower()=="profiles"):
#        plotProfiles          = bool(v)
#    pass
#    
#    if(confidenceLevel > 1.0 or confidenceLevel <= 0.0):
#      raise RuntimeError("Confidence level must be 0 < cl < 1. Ex. use 0.683 for 1-sigma interval")
#    
#    #Get chisq critical value corresponding to this confidence level
#    if(equivalentSigma==None):
#      equivalentSigma         = scipy.stats.norm.isf((1-confidenceLevel)/2.0) 
#    else:
#      confidenceLevel         = 1-(scipy.stats.norm.sf(equivalentSigma)*2.0)
#    pass
#    
#    criticalValue             = scipy.stats.chi2.isf([1-confidenceLevel],1)[0]
#    
#    print("Computing %.3f c.l. errors (chisq critical value: %.3f, equivalent sigmas: %.3f sigma)" %(confidenceLevel,criticalValue,equivalentSigma))
#    
#    #Now computing the errors
#    if(len(self.bestFitValues.keys())==0):
#      raise RuntimeError("You have to perform a fit before calling computeErrors!")
#    
#    #Find confidence intervals for all parameters, except nuisance ones
#    paramNames                = self.bestFitValues.keys()
#    paramList                 = []
#    for par in paramNames:
#      if(self.modelManager[par].isNuisance()):
#        continue
#      else:
#        paramList.append(par)
#      pass
#    pass
#    
#    confInterval              = collections.OrderedDict()
#    
#    for i,parname in enumerate(paramList):
#      
#      sys.stdout.write("Computing error for parameter %s...\n" %(parname))
#      
#      #Get the list of free parameters
#      self.freeParameters     = self.modelManager.getFreeParameters()
#      
#      self._restoreBestFitValues()
#      
#      #Remove the current parameter from the list of free parameters,
#      #so that it won't be varied
#      self.freeParameters.pop(parname)
#      
#      #Build the profile logLike for this parameter
#      
#      def thisProfileLikeRenorm(newValue):
#        
#        self._restoreBestFitValues()
#        
#        #Set the current parameter to its current value
#        
#        #newValue              = newValue[0]
#        self.modelManager[parname].setValue(newValue)
#        
#        #Fit all other parameters
#        
#        minimizer             = self.Minimizer(self.minusLogLikeProfile,self.freeParameters)
#        _,_,proflogL          = minimizer.minimize(False,False)
#        
#        #Subtract the minimum and the critical value/2, so that when this is 0 the true profileLogLike is 
#        #logL+critical value/2.0
#        #(the factor /2.0 comes from the LRT, which has 2*deltaLoglike as statistic)
#        
#        return proflogL-self.logLmin-criticalValue/2.0
#      pass
#      
#      #Find the values of the parameter for which the profile logLike is
#      # equal to the minimum - critical value/2.0, i.e., when thisProfileLikeRenorm is 0
#      
#      #We will use the approximate error (the sqrt of the diagonal of the covariance matrix)
#      #as starting point for the search. Since it represents the approximate 1 sigma error,
#      #we have to multiply it by the appropriate number of sigmas
#      
#      bounds                  = []
#      for kind in ['lower','upper']:
#        if(kind=='lower'):
#          for i in [1.0,1.1,1.5,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0]:
#            approxSolution        = self.bestFitValues[parname]-i*equivalentSigma*abs(self.approxErrors[parname])
#            if(thisProfileLikeRenorm(approxSolution) > 0):
#              break
#        else:
#          for i in [1.0,1.1,1.5,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0]:
#            approxSolution        = self.bestFitValues[parname]+i*equivalentSigma*abs(self.approxErrors[parname])
#            if(thisProfileLikeRenorm(approxSolution) > 0):
#              break
#        pass
#        
#        if(approxSolution < self.modelManager[parname].minValue):
#          approxSolution          = self.modelManager[parname].minValue*1.1
#        pass
#        
#        if(approxSolution > self.modelManager[parname].maxValue):
#          approxSolution          = self.modelManager[parname].maxValue*0.9
#        pass
#        
#        tolerance                 = abs(self.bestFitValues[parname])/10000.0
#        
#        if(self.verbose):
#          print("Approx solution for %s bound is %s, tolerance is %s" %(kind,approxSolution,tolerance))
#        try:
#          #This find the root of thisProfileLikeRenorm, i.e., the value of its argument for which
#          #it is zero
#          #results                 = scipy.optimize.root(thisProfileLikeRenorm,
#          #                                              approxSolution,
#          #                                              method='lm')
#          results                 = scipy.optimize.brentq(thisProfileLikeRenorm,approxSolution,self.bestFitValues[parname],rtol=1e-3)
#          
#        except:
#          print("Error search for %s bound for parameter %s failed. Parameter is probably unconstrained." %(kind,parname))
#          raise
#        else:
#          
#          #if(results['success']==False):
#          #  print RuntimeError("Could not find a solution for the %s bound confidence for parameter %s" %(kind,parname))
#          #  raise
#          #bounds.append(results['x'][0])
#          bounds.append(results) 
#      pass
#                  
#      confInterval[parname] = [min(bounds),max(bounds)]
#    pass
#    
#    self.freeParameters     = self.modelManager.getFreeParameters()
#    
#    print("\nBest fit values and their errors are:")
#    for parname in confInterval.keys():
#      value                   = self.bestFitValues[parname]
#      error1,error2           = confInterval[parname]
#      print("%-20s = %6.3g [%6.4g,%6.4g]" %(parname,value,error1-value,error2-value))
#    pass
#    
#    if(plotProfiles):
#      #Plot the profile likelihoods for each parameter
#      npar                    = len(confInterval.keys())
#      nrows                   = npar/2
#      ncols                   = 2
#      if(nrows*ncols < npar):
#        nrow                 += 1
#      pass
#      
#      fig,subs                = plt.subplots(nrows,ncols)
#      
#      for i,sub,(parname,interval) in zip(range(npar),subs.flatten(),confInterval.iteritems()):
#        #Remove this parameter from the freeParameters list
#        #Get the list of free parameters
#        self.freeParameters     = self.modelManager.getFreeParameters()
#      
#        self._restoreBestFitValues()
#        
#        #Remove the current parameter from the list of free parameters,
#        #so that it won't be varied
#        self.freeParameters.pop(parname)
#        
#        val                   = self.bestFitValues[parname]
#        errorM                = interval[0]-val
#        errorP                = interval[1]-val   
#        grid                  = numpy.linspace(val+1.1*errorM,val+1.1*errorP,10)
#        grid                  = numpy.append(grid,val)
#        grid.sort()
#        logLonGrid            = []
#        for g in grid:
#          self._restoreBestFitValues()
#          logLonGrid.append(2*(thisProfileLikeRenorm(g)+criticalValue/2.0))
#        pass
#        sub.plot(grid,logLonGrid)
#        sub.set_xlabel("%s" %(parname))
#        sub.set_ylabel(r"2 ($L_{prof}$-L$_{0}$)")
#        sub.axhline(criticalValue,linestyle='--')
#        #Reduce the number of x ticks
#        sub.locator_params(nbins=5,axis='x')
#      pass
#      
#      plt.tight_layout()
#      
#    pass
#    
#    return confInterval
#  pass




import collections
from Parameter import Parameter

from threeML.io.Table import Table

from IPython.display import display, Latex, HTML

class SourceModel(object):  
  
  def __init__(self):
    print("You should not see this!")
  
  def __getitem__(self,argument):
    return self.parameters[argument]
  
  def getAllParameters(self):
  
    allParameters                  = collections.OrderedDict()
    
    for dic in [self.parameters,self.spectralModel.parameters]:
      
      for k,v in dic.iteritems():
        
        allParameters[k]           = v
    
    return allParameters 
    
  
  def __repr__(self):
    
    print("Spatial model: %s" %(self.functionName))
    print("Formula:\n")
    
    display(Latex(self.formula))
    
    if(hasattr(self,'spectralModel')):
      print("\nSpectral model: %s" %(self.spectralModel.functionName))
    
    print("\nCurrent parameters:\n")
    
    data = []
    nameLength = 0
    
    for dic in [self.parameters,self.spectralModel.parameters]:
      for k,v in dic.iteritems():
        if(v.isFree()):
          ff                   = "free"
        else:
          ff                   = "fixed"
        pass
        data.append([v.name,v.value,v.minValue,v.maxValue,v.delta,ff,v.unit])
        
        if(len(v.name) > nameLength):
          nameLength = len(v.name)
      
      pass
    pass
    
    table                     = Table(rows = data,
                                      names = ["Name","Value","Minimum","Maximum","Delta","Status","Unit"],
                                      dtype=('S%i' %nameLength, float, float, float, float, "S5",'S15'))
    
    display(table)
        
    return ''
  pass
  
pass



from threeML.models.spectralmodel import SpectralModel
from threeML.models.Parameter import Parameter
import numpy
import math
import scipy.integrate
import operator
import collections


class Madness(SpectralModel):
  def setup(self,model,nBreaks=100,emin=1.0,emax=1e9,**kwargs):
    self.functionName        = "Madness"
    self.formula             = r'\begin{equation}f(E) = K_{0} E^{\alpha}\end{equation}'
    self.parameters          = collections.OrderedDict()
    
    self.nBreaks             = int(nBreaks)
    self.energyBreaks        = numpy.logspace(numpy.log10(emin),numpy.log10(emax),self.nBreaks+2)[1:-1]
    
    for k,v in kwargs.iteritems():
      if(k.lower()=='breaks'):
        print("Using user-provided breaks")
        self.energyBreaks    = numpy.array(v)
        self.nBreaks         = self.energyBreaks.shape[0]
      pass
    pass
    
    #Choose the pivot energies equal to the breaks, for convenience. Use emin as pivot energy
    #for the first branch, before the first break
    self.pivotEnergies       = numpy.concatenate([[emin],self.energyBreaks])
        
    #Initial values for alphas: a curved spectrum starting with -1
    #and ending with -2.5, like GRBs do when modeled with a Band function
    alphas                   = numpy.linspace(-2.5,-1,self.nBreaks+1)[::-1]
            
    #Add all the alphas to the parameters list
    for i in xrange(self.nBreaks+1):
      #Normalization parameter
      thisNorm                 = 'K_%s' %i
      pv                       = model(self.pivotEnergies[i])
      self.parameters[thisNorm] = Parameter(thisNorm,pv,pv/100,pv*100,0.1,fixed=False,nuisance=False,dataset=None,normalization=True)
      thisAlpha                = 'alpha_%s' %i
      self.parameters[thisAlpha]   = Parameter(thisAlpha,alphas[i],-10,10,0.1,fixed=True,nuisance=False,dataset=None)      
    pass
    
    def integral(e1,e2):
      #I'm too lazy now to write the analytical expression
      return self((e1+e2)/2.0)*(e2-e1)
      
    self.integral            = integral
    
    self.normalizations      = numpy.zeros(self.nBreaks+2)
    
    self._computeNormalizations()
  pass
  
  def _computeNormalizations(self):
    #Use a generator instead of a list to gain speed
    generator                = (x.value for x in self.parameters.values()[1::2])
    self.indexes             = numpy.fromiter(generator,float)
    indexDiff                = self.indexes[:-1]-self.indexes[1:]
              
    self.normalizations[0]   = self.parameters['K_0'].value
    self.normalizations[1:-1]  = (numpy.power(self.energyBreaks,indexDiff)*
                                  numpy.power(self.pivotEnergies[1:],self.indexes[1:])
                                  /
                                  numpy.power(self.pivotEnergies[:-1],self.indexes[:-1]))
    self.normalizations[-1]  = 1.0
    
    #This compute the cumulative product of the array
    #(i.e., the first elements is a0, the second a0*a1,
    #the third a0*a1*a2, and so on...)
    self.products            = numpy.cumprod(self.normalizations)
  pass
  
  def __call__(self,energy):
    self._computeNormalizations()
    
    #Make this always an array
    energies                 = numpy.array(energy,ndmin=1,copy=False)
    energies.sort()
    
    #This find the indexes of the places in which the breaks should be inserted
    #to keep the order of the array. In other words, for each break finds the index of
    #the first value larger or equal to the break energy
    indexes                  = numpy.searchsorted(energies,self.energyBreaks)
    indexes                  = numpy.concatenate(([0],indexes,[energies.shape[0]]))
    
    results                  = numpy.empty(energies.shape)
    
    for i in xrange(self.nBreaks+1):
      i1,i2                  = (indexes[i],indexes[i+1])
      thisNorm               = self.parameters['K_%s' %i].value
      thisPivot              = self.pivotEnergies[i]
      results[i1:i2]         = thisNorm*numpy.power(energies[i1:i2]/thisPivot,self.indexes[i])
    pass   
    
    return numpy.maximum(results,1e-30)
  pass
  
pass



import os, sys
import pkg_resources
import yaml


class Config( object ):
    
    def __init__( self ):
        
        # Read the config file
        
        # Define a list of possible path where the config file might be
        # The first successful path will be the active configuration file
        
        possiblePaths = []
        
        # First possible path is the .threeML directory under the user-home
        
        possiblePaths.append( os.path.join( os.path.expanduser( '~' ), '.threeML'  ) )
        
        # Second possible path is the config subdir under the package path
        # (which is where this Config.py file is)

        distribution = pkg_resources.get_distribution("threeML")

        possiblePaths.append( os.path.join( distribution.location, 'threeML/config' ) )

        self._configuration = None
        self._filename = None
        
        for path in possiblePaths:
            
            thisFilename = os.path.join( path, 'threeML_config.yml' )
            
            if os.path.exists( thisFilename ):
                
                with open( thisFilename ) as f:
                
                    self._configuration = yaml.safe_load( f )
                
                print("Configuration read from %s" %( thisFilename ) )
                
                self._filename = thisFilename
                
                break
            
            else:
                
                continue
        
        if self._configuration is None:
            
            raise RuntimeError("Could not find threeML_config.yml in any of %s" %( possiblePaths ))

    
    def __getitem__(self, key):
        
        if key in self._configuration.keys():
            
            return self._configuration[ key ]
        
        else:
            
            raise RuntimeError("Configuration key %s does not exist in %s." %( key, self._filename ) )
    
    def __repr__(self):
        
        return yaml.dump( self._configuration, default_flow_style=False )


# Now read the config file, so it will be available as Config.c
threeML_config = Config()




from band import Band
from scipy import weave
from scipy.weave import converters

class wBand(Band):
    def __call__(self,e):
      energies                 = numpy.array(e,ndmin=1,copy=False)
      alpha                    = self.parameters['alpha'].value
      beta                     = self.parameters['beta'].value
      E0                       = self.parameters['E0'].value
      K                        = self.parameters['K'].value
      
      n                        = energies.shape[0]
    
      code="""
        double *result = (double *)malloc(n*sizeof(double));
        int i;
        double ee,E,E100,ee100;
        ee = (alpha+beta)*E0;
        ee100 = ee/1000.0;
        for(i=0;i<n;i++)
        {
        E = energies(i);
        E100 = E/1000.0;
        if(E < ee) {
        result[i] = K*pow(E100,alpha)*exp(-E/E0);
        } else {
        result[i] = K*pow(ee100,alpha-beta)*exp(beta-alpha)*pow(E100,beta);
        }
        }
        return_val = result;
        """
      return weave.inline(code,['energies','alpha','beta','E0','K','n'],type_converters=converters.blitz)
  
 
